# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


import json
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Iterable, Mapping
from typing import Any

import pandas as pd
from ax.api.types import TParameterization

from ax.core.base_trial import BaseTrial, TrialStatus
from ax.core.map_data import MAP_KEY, MapData
from ax.core.map_metric import MapMetric
from ax.core.metric import MetricFetchE, MetricFetchResult
from ax.core.runner import Runner
from ax.core.trial import Trial
from ax.exceptions.storage import JSONEncodeError
from ax.utils.common.result import Err, Ok
from pyre_extensions import assert_is_instance, none_throws, override


class _APIMetric(MapMetric, ABC):
    """
    This Metric provides implmementations for the essential MapMetric methods expected
    by the Ax Orchestrator, shimming between our IMetric and MapMetric interfaces.

    Users should never instantiate or subclass this class directly.

    Ideally we will be able to remove this class in the future, once we have stablized
    structure of ax.core.Metric to be more in line with our long term vision for Ax.
    """

    def __init__(self, name: str) -> None:
        super().__init__(name=name)

    @abstractmethod
    def fetch(
        self,
        trial_index: int,
        trial_metadata: Mapping[str, Any],
    ) -> tuple[int, float | tuple[float, float]]: ...

    @override
    def fetch_trial_data(self, trial: BaseTrial, **kwargs: Any) -> MetricFetchResult:
        """
        Fetch data for a single trial for this metric by calling the user-implemented
        fetch method.
        """
        trial = assert_is_instance(trial, Trial)

        try:
            progression, outcome = self.fetch(
                trial_index=trial.index, trial_metadata=trial.run_metadata
            )

            if isinstance(outcome, float):
                mean = outcome
                sem = float("nan")
            else:
                mean, sem = outcome

            record = {
                "trial_index": trial.index,
                "arm_name": none_throws(trial.arm).name,
                "metric_name": self.name,
                MAP_KEY: progression,
                "mean": mean,
                "sem": sem,
            }
            return Ok(value=MapData(df=pd.DataFrame.from_records([record])))
        except Exception as e:
            return Err(
                value=MetricFetchE(message=f"Failed to fetch {self.name}", exception=e)
            )


class _APIRunner(Runner, ABC):
    """
    This Runner provides implementations for the essential Runner methods expected
    by the Ax Orchestrator, shimming between our IRunner and Runner interfaces.

    Users should never instantiate or subclass this class directly.

    Ideally we will be able to remove this class in the future, once we have stablized
    structure of ax.core.Runner to be more in line with our long term vision for Ax.
    """

    @abstractmethod
    def run_trial(
        self, trial_index: int, parameterization: TParameterization
    ) -> dict[str, Any]: ...

    @abstractmethod
    def poll_trial(
        self, trial_index: int, trial_metadata: Mapping[str, Any]
    ) -> TrialStatus: ...

    @abstractmethod
    def stop_trial(
        self, trial_index: int, trial_metadata: Mapping[str, Any]
    ) -> dict[str, Any]: ...

    @override
    def run(self, trial: BaseTrial) -> dict[str, Any]:
        """
        Runs a trial by calling the user-implemented run_trial method, returning
        appropriate metadata.
        """
        metadata = self.run_trial(
            trial_index=trial.index,
            # pyre-ignore[6] Arms in core Ax may have None in their parameters
            parameterization=none_throws(
                assert_is_instance(trial, Trial).arm
            ).parameters,
        )

        # Runtime validate metadata is JSON serializable to avoid issues when
        # serializing the Trial.
        try:
            json.dumps(metadata)
            return metadata
        except TypeError:
            raise JSONEncodeError(
                f"Metadata must be JSON serializable, received {metadata}"
            )

    @override
    def poll_trial_status(
        self, trials: Iterable[BaseTrial]
    ) -> dict[TrialStatus, set[int]]:
        """
        Polls the status of trials by calling the user-implemented poll_trial method.
        """
        res = defaultdict(set)
        for trial in trials:
            status = self.poll_trial(
                trial_index=trial.index, trial_metadata=trial.run_metadata
            )

            res[status].add(trial.index)

        return res

    @override
    def stop(self, trial: BaseTrial, reason: str | None = None) -> dict[str, Any]:
        """Stops a trial by calling the user-implemented stop_trial method"""
        return self.stop_trial(
            trial_index=trial.index, trial_metadata=trial.run_metadata
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from logging import Logger

from ax.core.experiment import Experiment
from ax.generation_strategy.generation_strategy import GenerationStrategy
from ax.storage.sqa_store.db import session_scope
from ax.storage.sqa_store.decoder import Decoder
from ax.storage.sqa_store.sqa_classes import SQAExperiment
from ax.storage.sqa_store.sqa_config import SQAConfig
from ax.utils.common.logger import get_logger

logger: Logger = get_logger(__name__)


def delete_experiment(exp_name: str) -> None:
    """Delete experiment by name.

    Args:
        experiment_name: Name of the experiment to delete.
    """
    with session_scope() as session:
        exp = session.query(SQAExperiment).filter_by(name=exp_name).one_or_none()
        session.delete(exp)
        session.flush()

    logger.warning(
        f"You are deleting {exp_name} and all its associated data from the database."
    )


def delete_generation_strategy(
    exp_name: str, config: SQAConfig | None = None, max_gs_to_delete: int = 1
) -> None:
    """Delete the generation strategy associated with an experiment

    Warning: This method will delete all existing generator runs and arms from
        the experiment. Use with caution.

    Args:
        exp_name: Name of the experiment for which the generation strategy
            should be deleted.
        config: The SQAConfig.
        max_gs_to_delete: There is never supposed to be more than one generation
            strategy associated with an experiment. However, we've seen cases where
            there are, and we don't know why. This parameter allows us to delete
            multiple generation strategies, but we raise an error if there are more
            than `max_gs_to_delete` generation strategies associated with the
            experiment.
            This is a safeguard in case we have a bug in this code that deletes
            all generation strategies.

    """
    config = config or SQAConfig()
    decoder = Decoder(config=config)
    exp_sqa_class = decoder.config.class_to_sqa_class[Experiment]
    gs_sqa_class = decoder.config.class_to_sqa_class[GenerationStrategy]
    # get the generation strategy's db_id
    with session_scope() as session:
        sqa_gs_ids = (
            session.query(gs_sqa_class.id)  # pyre-ignore[16]
            .join(exp_sqa_class.generation_strategy)  # pyre-ignore[16]
            # pyre-fixme[16]: `SQABase` has no attribute `name`.
            .filter(exp_sqa_class.name == exp_name)
            .all()
        )

    if sqa_gs_ids is None:
        logger.debug(f"No generation strategy found for experiment {exp_name}.")
        return None

    if len(sqa_gs_ids) > max_gs_to_delete:
        raise ValueError(
            f"Found {len(sqa_gs_ids)} generation strategies for experiment {exp_name}. "
            "If you are sure you want to delete all of them, please set "
            f"`max_gs_to_delete` (currently {max_gs_to_delete}) to a higher value."
        )

    # delete generation strategy
    with session_scope() as session:
        gs_list = (
            session.query(gs_sqa_class)
            .filter(gs_sqa_class.id.in_([id[0] for id in sqa_gs_ids]))
            .all()
        )
        for gs in gs_list:
            session.delete(gs)
        session.flush()



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from logging import Logger
from typing import Optional, TYPE_CHECKING

from ax.adapter.data_utils import ExperimentData
from ax.adapter.transforms.base import Transform
from ax.core.observation import Observation, ObservationFeatures
from ax.core.parameter import ChoiceParameter, ParameterType
from ax.core.search_space import RobustSearchSpace, SearchSpace
from ax.core.utils import get_target_trial_index
from ax.exceptions.core import UnsupportedError
from ax.generators.types import TConfig
from ax.utils.common.logger import get_logger
from pyre_extensions import none_throws

if TYPE_CHECKING:
    # import as module to make sphinx-autodoc-typehints happy
    from ax import adapter as adapter_module  # noqa F401


TRIAL_PARAM = "TRIAL_PARAM"
logger: Logger = get_logger(__name__)


class TrialAsTask(Transform):
    """Convert trial to one or more task parameters.

    How trial is mapped to parameter is specified with a map like
    {parameter_name: {trial_index: level name}}.
    For example,
    {"trial_param1": {0: "level1", 1: "level1", 2: "level2"},}
    will create choice parameters "trial_param1" with is_task=True.
    Observations with trial 0 or 1 will have "trial_param1" set to "level1",
    and those with trial 2 will have "trial_param1" set to "level2". Multiple
    parameter names and mappings can be specified in this dict.

    The trial level mapping can be specified in config["trial_level_map"]. If
    not specified, defaults to a parameter with a level for every trial index.

    For the reverse transform, if there are multiple mappings in the transform
    the trial will not be set.

    The created parameter will be given a target value that will default to the
    lowest trial index in the mapping, or can be provided in config["target_trial"].

    Will raise if trial not specified for every point in the training data.

    If there are fewer than 2 trials or levels, the transform is a no-op.

    Transform is done in-place.
    """

    requires_data_for_initialization: bool = True

    def __init__(
        self,
        search_space: SearchSpace | None = None,
        observations: list[Observation] | None = None,
        experiment_data: ExperimentData | None = None,
        adapter: Optional["adapter_module.base.Adapter"] = None,
        config: TConfig | None = None,
    ) -> None:
        super().__init__(
            search_space=search_space,
            observations=observations,
            experiment_data=experiment_data,
            adapter=adapter,
            config=config,
        )
        assert adapter is not None, "TrialAsTask requires adapter"
        if isinstance(search_space, RobustSearchSpace):
            raise UnsupportedError(
                "TrialAsTask transform is not supported for RobustSearchSpace."
            )
        # Identify values of trial.
        if experiment_data is not None:
            trials = set(experiment_data.arm_data.index.get_level_values("trial_index"))
        else:
            trials = {obs.features.trial_index for obs in none_throws(observations)}
        if None in trials:
            raise ValueError(
                "Unable to use trial as task since not all observations have "
                "trial specified."
            )
        # Get trial level map
        if config is not None and "trial_level_map" in config:
            # pyre-ignore [9]
            trial_level_map: dict[str, dict[int | str, int | str]] = config[
                "trial_level_map"
            ]
            # Validate
            self.trial_level_map: dict[str, dict[int, int | str]] = {}
            for _p_name, level_dict in trial_level_map.items():
                # cast trial index as an integer
                int_keyed_level_dict = {
                    int(trial_index): v for trial_index, v in level_dict.items()
                }
                self.trial_level_map[_p_name] = int_keyed_level_dict
                # Check that trials match those in data
                level_map = set(int_keyed_level_dict.keys())
                if not trials.issubset(level_map):
                    raise ValueError(
                        f"Not all trials in data ({trials}) contained "
                        f"in trial level map for {_p_name} ({level_map})"
                    )
        else:
            # Set TRIAL_PARAM for each trial to the corresponding trial_index.
            self.trial_level_map = {TRIAL_PARAM: {int(b): str(b) for b in trials}}
        if len(self.trial_level_map) == 1:
            level_dict = next(iter(self.trial_level_map.values()))
            self.inverse_map: dict[int | str, int] | None = {
                v: k for k, v in level_dict.items()
            }
        else:
            self.inverse_map = None
        # Compute target values
        self.target_values: dict[str, int | str] = {}

        for p_name, trial_map in list(self.trial_level_map.items()):
            if len(set(trial_map.values())) < 2:
                # If there are less than two distinct levels, then we don't need to
                # create a task parameter and the transform is a no-op.
                del self.trial_level_map[p_name]
                continue
            if config is not None and "target_trial" in config:
                target_trial = int(config["target_trial"])
            else:
                target_trial = none_throws(
                    get_target_trial_index(
                        experiment=none_throws(none_throws(adapter)._experiment)
                    )
                )
                logger.debug(f"Setting target value for {p_name} to {target_trial}")
            self.target_values[p_name] = trial_map[target_trial]

    def transform_observation_features(
        self, observation_features: list[ObservationFeatures]
    ) -> list[ObservationFeatures]:
        """If the trial_index is in ``level_dict``, which is from trials with data,
        use the value for that trial index from ``level_dict``, which is probably
        the stringified trial index, as the task parameter. If the trial index is
        None or not in ``level_dict``, use the value from the most recent trial.

        Trial indices not observed are probably pending points from candidate or staged
        trials. Trial indices set to None are probably pending points passed in by the
        user.
        """
        if len(self.trial_level_map) == 0:
            # no-op
            return observation_features
        for obsf in observation_features:
            for p_name, level_dict in self.trial_level_map.items():
                if obsf.trial_index is not None and int(obsf.trial_index) in level_dict:
                    obsf.parameters[p_name] = level_dict[int(obsf.trial_index)]
                elif len(obsf.parameters) > 0:
                    # If the trial index is none, but the parameters are not empty
                    # perform the transform by assuming the observation is from the
                    # most recent trial. This is needed for generating trials composed
                    # of points from multiple models.
                    obsf.parameters[p_name] = level_dict[max(level_dict)]
            obsf.trial_index = None
        return observation_features

    def _transform_search_space(self, search_space: SearchSpace) -> SearchSpace:
        if len(self.trial_level_map) == 0:
            # no-op
            return search_space
        for p_name, level_dict in self.trial_level_map.items():
            level_values = sorted(set(level_dict.values()))
            if len(level_values) < 2:
                details = (
                    f"only 1 found: {level_values}" if level_values else "none found"
                )
                raise ValueError(
                    f"TrialAsTask transform expects 2+ task params, {details}"
                )
            is_int = all(isinstance(val, int) for val in level_values)
            trial_param = ChoiceParameter(
                name=p_name,
                parameter_type=ParameterType.INT if is_int else ParameterType.STRING,
                values=level_values,  # pyre-fixme [6]
                # if all values are integers, retain the original order
                # they are encoded in TaskChoiceToIntTaskChoice
                is_ordered=is_int,
                is_task=True,
                sort_values=True,
                target_value=self.target_values[p_name],
            )
            search_space.add_parameter(trial_param)
        return search_space

    def untransform_observation_features(
        self, observation_features: list[ObservationFeatures]
    ) -> list[ObservationFeatures]:
        """If task parameters have been added to observation features by
        this parameter, then remove those task parameters and restore
        the trial index/

        Args:
            observation_features: List of observation features to untransform.

        Returns:
            List of observation features with task parameters removed and trial
                index restored.
        """
        if len(self.trial_level_map) == 0:
            # no-op
            return observation_features
        for obsf in observation_features:
            for p_name in self.trial_level_map:
                pval = obsf.parameters.pop(p_name)
            # pyre-fixme[61]: `pval` may not be initialized here.
            if self.inverse_map is not None and pval in self.inverse_map:
                # pyre-fixme[61]: `pval` may not be initialized here.
                obsf.trial_index = self.inverse_map[pval]
        return observation_features

    def transform_experiment_data(
        self, experiment_data: ExperimentData
    ) -> ExperimentData:
        arm_data = experiment_data.arm_data
        for p_name, level_dict in self.trial_level_map.items():
            arm_data[p_name] = arm_data.index.get_level_values("trial_index").map(
                level_dict
            )
        return ExperimentData(
            arm_data=arm_data, observation_data=experiment_data.observation_data
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from typing import Optional, TYPE_CHECKING

from ax.adapter.data_utils import ExperimentData
from ax.adapter.transforms.base import Transform
from ax.adapter.transforms.utils import construct_new_search_space
from ax.core.observation import Observation, ObservationFeatures
from ax.core.parameter import (
    ChoiceParameter,
    DerivedParameter,
    FixedParameter,
    RangeParameter,
)
from ax.core.search_space import SearchSpace
from ax.generators.types import TConfig

if TYPE_CHECKING:
    # import as module to make sphinx-autodoc-typehints happy
    from ax import adapter as adapter_module  # noqa F401


class RemoveFixed(Transform):
    """Remove fixed and derived parameters.

    Fixed and derived parameters should not be included in the SearchSpace.
    This transform removes these parameters, leaving only tunable parameters.

    Transform is done in-place for observation features.
    """

    def __init__(
        self,
        search_space: SearchSpace | None = None,
        observations: list[Observation] | None = None,
        experiment_data: ExperimentData | None = None,
        adapter: Optional["adapter_module.base.Adapter"] = None,
        config: TConfig | None = None,
    ) -> None:
        assert search_space is not None, "RemoveFixed requires search space"
        super().__init__(
            search_space=search_space,
            observations=observations,
            experiment_data=experiment_data,
            adapter=adapter,
            config=config,
        )
        # Identify parameters that should be transformed
        self.fixed_or_derived_parameters: dict[
            str, FixedParameter | DerivedParameter
        ] = {
            p_name: p
            for p_name, p in search_space.parameters.items()
            if isinstance(p, (DerivedParameter, FixedParameter))
        }

    def transform_observation_features(
        self, observation_features: list[ObservationFeatures]
    ) -> list[ObservationFeatures]:
        for obsf in observation_features:
            for p_name in self.fixed_or_derived_parameters:
                obsf.parameters.pop(p_name, None)
        return observation_features

    def _transform_search_space(self, search_space: SearchSpace) -> SearchSpace:
        tunable_parameters: list[ChoiceParameter | RangeParameter] = []
        for p in search_space.parameters.values():
            if p.name not in self.fixed_or_derived_parameters:
                # If it's not in fixed_or_derived_parameters, it must be a
                # tunable param.
                # pyre: p_ is declared to have type `Union[ChoiceParameter,
                # pyre: RangeParameter]` but is used as type `ax.core.
                # pyre-fixme[9]: parameter.Parameter`.
                p_: ChoiceParameter | RangeParameter = p
                tunable_parameters.append(p_)
        return construct_new_search_space(
            search_space=search_space,
            # pyre-ignore Incompatible parameter type [6]
            parameters=tunable_parameters,
            parameter_constraints=[
                pc.clone() for pc in search_space.parameter_constraints
            ],
        )

    def untransform_observation_features(
        self, observation_features: list[ObservationFeatures]
    ) -> list[ObservationFeatures]:
        for obsf in observation_features:
            for p_name, p in self.fixed_or_derived_parameters.items():
                if isinstance(p, DerivedParameter):
                    obsf.parameters[p_name] = p.compute(parameters=obsf.parameters)
                else:
                    obsf.parameters[p_name] = p.value
        return observation_features

    def transform_experiment_data(
        self, experiment_data: ExperimentData
    ) -> ExperimentData:
        return ExperimentData(
            arm_data=experiment_data.arm_data.drop(
                columns=list(self.fixed_or_derived_parameters)
            ),
            observation_data=experiment_data.observation_data,
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


import copy
import math
from typing import cast
from unittest.mock import patch, PropertyMock

import torch
from ax.adapter.base import Adapter
from ax.adapter.registry import Generators
from ax.adapter.torch import TorchAdapter
from ax.generators.torch.botorch import LegacyBoTorchGenerator
from ax.utils.common.random import set_rng_seed
from ax.utils.common.testutils import TestCase
from ax.utils.sensitivity.derivative_gp import posterior_derivative
from ax.utils.sensitivity.derivative_measures import (
    compute_derivatives_from_model_list,
    GpDGSMGpMean,
    GpDGSMGpSampling,
    sample_discrete_parameters,
)
from ax.utils.sensitivity.sobol_measures import (
    _get_model_per_metric,
    ax_parameter_sens,
    compute_sobol_indices_from_model_list,
    ProbitLinkMean,
    SobolSensitivity,
    SobolSensitivityGPMean,
    SobolSensitivityGPSampling,
)
from ax.utils.testing.core_stubs import get_branin_experiment
from ax.utils.testing.mock import mock_botorch_optimize
from botorch.models.gpytorch import BatchedMultiOutputGPyTorchModel, GPyTorchModel
from botorch.models.model_list_gp_regression import ModelListGP
from botorch.utils.transforms import unnormalize
from gpytorch.distributions import MultivariateNormal
from torch import Tensor


@mock_botorch_optimize
def get_adapter(modular: bool = False, saasbo: bool = False) -> Adapter:
    exp = get_branin_experiment(with_batch=True)
    exp.trials[0].run()
    if modular:
        return Generators.BOTORCH_MODULAR(
            experiment=exp,
            data=exp.fetch_data(),
        )
    if saasbo:
        return Generators.SAASBO(
            experiment=exp,
            data=exp.fetch_data(),
        )
    else:
        return Generators.LEGACY_BOTORCH(experiment=exp, data=exp.fetch_data())


class SensitivityAnalysisTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        torch.manual_seed(0)
        self.model = get_adapter().generator.model
        self.saas_model = get_adapter(saasbo=True).generator.surrogate.model

    def test_DgsmGpMean(self) -> None:
        bounds = torch.tensor([(0.0, 1.0) for _ in range(2)]).t()
        sensitivity_mean = GpDGSMGpMean(self.model, bounds=bounds, num_mc_samples=10)
        gradients_measure = sensitivity_mean.gradient_measure()
        gradients_absolute_measure = sensitivity_mean.gradient_absolute_measure()
        gradients_square_measure = sensitivity_mean.gradients_square_measure()
        self.assertIsInstance(gradients_measure, Tensor)
        self.assertIsInstance(gradients_absolute_measure, Tensor)
        self.assertIsInstance(gradients_square_measure, Tensor)
        self.assertEqual(gradients_measure.shape, torch.Size([2]))
        self.assertEqual(gradients_absolute_measure.shape, torch.Size([2]))
        self.assertEqual(gradients_square_measure.shape, torch.Size([2]))

        sensitivity_mean_bootstrap = GpDGSMGpMean(
            self.model, bounds=bounds, num_mc_samples=10, num_bootstrap_samples=10
        )
        gradients_measure = sensitivity_mean_bootstrap.gradient_measure()
        gradients_absolute_measure = (
            sensitivity_mean_bootstrap.gradient_absolute_measure()
        )
        gradients_square_measure = sensitivity_mean_bootstrap.gradients_square_measure()
        self.assertIsInstance(gradients_measure, Tensor)
        self.assertIsInstance(gradients_absolute_measure, Tensor)
        self.assertIsInstance(gradients_square_measure, Tensor)
        self.assertEqual(gradients_measure.shape, torch.Size([2, 3]))
        self.assertEqual(gradients_absolute_measure.shape, torch.Size([2, 3]))
        self.assertEqual(gradients_square_measure.shape, torch.Size([2, 3]))

        # Test derivatives from model list
        res = compute_derivatives_from_model_list(
            [self.model for _ in range(2)],
            bounds=bounds,
        )
        self.assertEqual(res.shape, (2, 2))

    def test_DgsmGpSampling(self) -> None:
        bounds = torch.tensor([(0.0, 1.0) for _ in range(2)]).t()
        sensitivity_sampling = GpDGSMGpSampling(
            self.model, bounds=bounds, num_mc_samples=10, num_gp_samples=10
        )
        gradients_measure = sensitivity_sampling.gradient_measure()
        gradients_absolute_measure = sensitivity_sampling.gradient_absolute_measure()
        gradients_square_measure = sensitivity_sampling.gradients_square_measure()
        self.assertIsInstance(gradients_measure, Tensor)
        self.assertIsInstance(gradients_absolute_measure, Tensor)
        self.assertIsInstance(gradients_square_measure, Tensor)
        self.assertEqual(gradients_measure.shape, torch.Size([2, 3]))
        self.assertEqual(gradients_absolute_measure.shape, torch.Size([2, 3]))
        self.assertEqual(gradients_square_measure.shape, torch.Size([2, 3]))

        sensitivity_sampling_bootstrap = GpDGSMGpSampling(
            self.model,
            bounds=bounds,
            num_mc_samples=10,
            num_gp_samples=10,
            num_bootstrap_samples=10,
        )
        gradients_measure = sensitivity_sampling_bootstrap.gradient_measure()
        gradients_absolute_measure = (
            sensitivity_sampling_bootstrap.gradient_absolute_measure()
        )
        gradients_square_measure = (
            sensitivity_sampling_bootstrap.gradients_square_measure()
        )
        self.assertIsInstance(gradients_measure, Tensor)
        self.assertIsInstance(gradients_absolute_measure, Tensor)
        self.assertIsInstance(gradients_square_measure, Tensor)
        self.assertEqual(gradients_measure.shape, torch.Size([2, 5]))
        self.assertEqual(gradients_absolute_measure.shape, torch.Size([2, 5]))
        self.assertEqual(gradients_square_measure.shape, torch.Size([2, 5]))

    def _test_sobol_gp_mean(
        self,
        sensitivity: SobolSensitivityGPMean,
        expected_first_order: Tensor,
        expected_total_order: Tensor,
        expected_second_order: Tensor | None = None,
    ) -> None:
        """
        Check that outputs are as expected. The `assertAllClose` checks are
        characterization tests rather than correctness tests; they check that
        behavior is the same as it was in the past, not that it is
        quantitatively correct. Innocuous changes such as changing a random seed
        could potentially break these tests, and it may become necessary to
        delete them.
        """
        atol = 4e-3
        rtol = 2e-3
        first_order = sensitivity.first_order_indices()
        self.assertIsInstance(first_order, Tensor)
        self.assertAllClose(first_order, expected_first_order, atol=atol, rtol=rtol)
        self.assertEqual(first_order.shape, expected_first_order.shape)

        total_order = sensitivity.total_order_indices()
        self.assertIsInstance(total_order, Tensor)
        self.assertAllClose(total_order, expected_total_order, atol=atol, rtol=rtol)
        self.assertEqual(total_order.shape, expected_total_order.shape)

        if expected_second_order is not None:
            second_order = sensitivity.second_order_indices()
            self.assertIsInstance(second_order, Tensor)
            self.assertAllClose(
                second_order, expected_second_order, atol=atol, rtol=rtol
            )
            self.assertEqual(second_order.shape, expected_second_order.shape)

    def test_SobolGPMean(self) -> None:
        bounds = torch.tensor([(0.0, 1.0) for _ in range(2)]).t()
        sensitivity_mean = SobolSensitivityGPMean(
            self.model, num_mc_samples=10, bounds=bounds, second_order=True
        )
        self._test_sobol_gp_mean(
            sensitivity=sensitivity_mean,
            expected_first_order=torch.tensor(
                [-0.463695, -0.359848], dtype=torch.float64
            ),
            expected_total_order=torch.tensor(
                [1.015421, 1.052340], dtype=torch.float64
            ),
            expected_second_order=torch.tensor([0.823334], dtype=torch.float64),
        )

    def test_SobolGPMean_SAASBO(self) -> None:
        bounds = torch.tensor([(0.0, 1.0) for _ in range(2)]).t()
        sensitivity_mean_saas = SobolSensitivityGPMean(
            self.saas_model, num_mc_samples=10, bounds=bounds, second_order=True
        )
        self._test_sobol_gp_mean(
            sensitivity=sensitivity_mean_saas,
            expected_first_order=torch.tensor([0.5752, 0.5143], dtype=torch.double),
            expected_total_order=torch.tensor([0.9897, 0.0979], dtype=torch.float64),
            expected_second_order=torch.tensor([0.8300], dtype=torch.double),
        )

        sensitivity_mean_bootstrap = SobolSensitivityGPMean(
            self.model,
            num_mc_samples=10,
            bounds=bounds,
            second_order=True,
            num_bootstrap_samples=10,
            input_qmc=True,
        )
        self._test_sobol_gp_mean(
            sensitivity=sensitivity_mean_bootstrap,
            expected_first_order=torch.tensor(
                [[0.5511, 0.0237, 0.0487], [0.0878, 0.2175, 0.1475]],
                dtype=torch.float64,
            ),
            expected_total_order=torch.tensor(
                [[0.6946, 0.0242, 0.0497], [0.8412, 0.1112, 0.1062]],
                dtype=torch.float64,
            ),
            expected_second_order=torch.tensor(
                [[0.3119, 0.5284, 0.2299]], dtype=torch.float64
            ),
        )

        sensitivity_mean_bootstrap = SobolSensitivityGPMean(
            self.model,
            num_mc_samples=10,
            bounds=bounds,
            second_order=True,
            num_bootstrap_samples=10,
            link_function=ProbitLinkMean,
        )
        self._test_sobol_gp_mean(
            sensitivity=sensitivity_mean_bootstrap,
            expected_first_order=torch.tensor(
                [[0.4812, 0.1397, 0.1182], [1.9419, 2.7505, 0.5245]],
                dtype=torch.float64,
            ),
            expected_total_order=torch.tensor(
                [[0.2238, 0.0941, 0.0970], [0.6736, 0.2878, 0.1697]],
                dtype=torch.float64,
            ),
            expected_second_order=torch.tensor(
                [[0.8344, 7.9860, 0.8936]], dtype=torch.float64
            ),
        )

        sensitivity_mean = SobolSensitivityGPMean(
            self.model, num_mc_samples=10, bounds=bounds, second_order=False
        )
        self._test_sobol_gp_mean(
            sensitivity=sensitivity_mean,
            expected_first_order=torch.tensor([-0.0408, 0.4454], dtype=torch.float64),
            expected_total_order=torch.tensor([0.4405, 0.6326], dtype=torch.float64),
        )

        with self.assertRaisesRegex(ValueError, "Second order indices"):
            sensitivity_mean.second_order_indices()

        # testing compute_sobol_indices_from_model_list
        num_models = 3
        num_mc_samples = 10
        for order in ["first", "total", "second"]:
            with self.subTest(order=order):
                indices = compute_sobol_indices_from_model_list(
                    [self.model for _ in range(num_models)],
                    bounds=bounds,
                    order=order,
                    num_mc_samples=num_mc_samples,
                    input_qmc=True,
                )
                if order == "second":
                    self.assertEqual(indices.shape, (num_models, 1))
                else:
                    self.assertEqual(indices.shape, (num_models, 2))
                if order == "total":
                    self.assertTrue((indices >= 0).all())

                sobol_gp_mean = SobolSensitivityGPMean(
                    self.model,
                    bounds=bounds,
                    num_mc_samples=num_mc_samples,
                    input_qmc=True,
                    second_order=order == "second",
                )
                base_indices = getattr(sobol_gp_mean, f"{order}_order_indices")()
                # can compare values because we sample with deterministic seeds
                self.assertAllClose(
                    indices,
                    base_indices.unsqueeze(0).expand(num_models, indices.shape[1]),
                )

    def test_SobolGPMean_SAASBO_Ax_utils(self) -> None:
        num_mc_samples = 10
        for modular in [False, True]:
            adapter = cast(TorchAdapter, get_adapter(modular=modular))
            with self.assertRaisesRegex(
                NotImplementedError,
                "but only TorchAdapter is supported",
            ):
                # pyre-ignore
                ax_parameter_sens(1, adapter.outcomes)

            with patch.object(adapter, "generator", return_value=None):
                with self.assertRaisesRegex(
                    NotImplementedError,
                    "but only LegacyBoTorchGenerator and ModularBoTorchGenerator",
                ):
                    ax_parameter_sens(adapter, adapter.outcomes)

            torch_model = cast(LegacyBoTorchGenerator, adapter.generator)
            if not modular:
                with self.assertRaisesRegex(
                    NotImplementedError,
                    "but only ModelList is supported",
                ):
                    # only applies if the number of outputs of model is greater than 1
                    with patch.object(
                        BatchedMultiOutputGPyTorchModel,
                        "num_outputs",
                        new_callable=PropertyMock,
                    ) as mock:
                        mock.return_value = 2
                        ax_parameter_sens(adapter, adapter.outcomes)

                # since only ModelList is supported for LegacyBoTorchGenerator:
                gpytorch_model = ModelListGP(cast(GPyTorchModel, torch_model.model))
                torch_model.model = gpytorch_model

            for order in ["first", "total"]:
                with self.subTest(order=order):
                    ind_dict = ax_parameter_sens(
                        adapter,
                        input_qmc=True,
                        num_mc_samples=num_mc_samples,
                        order=order,
                        signed=False,
                    )
                    self.assertIsInstance(ind_dict, dict)

                    ind_tnsr = compute_sobol_indices_from_model_list(
                        _get_model_per_metric(torch_model, adapter.outcomes),
                        torch.tensor(torch_model.search_space_digest.bounds).T,
                        input_qmc=True,
                        num_mc_samples=num_mc_samples,
                        order=order,
                    )
                    self.assertIsInstance(ind_tnsr, Tensor)

                    # can compare values because we sample with deterministic seeds
                    for i, row in enumerate(ind_dict):
                        for j, col in enumerate(ind_dict[row]):
                            self.assertAlmostEqual(
                                ind_dict[row][col],
                                # pyre-fixme[6]: For 2nd argument expected
                                #  `SupportsRSub[Variable[_T],
                                #  SupportsAbs[SupportsRound[object]]]` but got
                                #  `Union[bool, float, int]`.
                                ind_tnsr[i, j].item(),
                            )
            with self.subTest(order="second"):
                second_ind_dict = ax_parameter_sens(
                    adapter,
                    input_qmc=True,
                    num_mc_samples=num_mc_samples,
                    order="second",
                    signed=False,
                )

                so_ind_tnsr = compute_sobol_indices_from_model_list(
                    _get_model_per_metric(torch_model, adapter.outcomes),
                    torch.tensor(torch_model.search_space_digest.bounds).T,
                    input_qmc=True,
                    num_mc_samples=num_mc_samples,
                    order="second",
                )
                fo_ind_tnsr = compute_sobol_indices_from_model_list(
                    _get_model_per_metric(torch_model, adapter.outcomes),
                    torch.tensor(torch_model.search_space_digest.bounds).T,
                    input_qmc=True,
                    num_mc_samples=num_mc_samples,
                    order="first",
                )
                # check that the first and second order indices are the same
                self.assertAlmostEqual(
                    second_ind_dict["branin"]["x1"],
                    # pyre-fixme[6]: For 2nd argument expected
                    #  `SupportsRSub[Variable[_T], SupportsAbs[SupportsRound[object]]]`
                    #  but got `Union[bool, float, int]`.
                    fo_ind_tnsr[0, 0].item(),
                )
                self.assertAlmostEqual(
                    second_ind_dict["branin"]["x2"],
                    # pyre-fixme[6]: For 2nd argument expected
                    #  `SupportsRSub[Variable[_T], SupportsAbs[SupportsRound[object]]]`
                    #  but got `Union[bool, float, int]`.
                    fo_ind_tnsr[0, -1].item(),
                )
                self.assertAlmostEqual(
                    second_ind_dict["branin"]["x1 & x2"],
                    # pyre-fixme[6]: For 2nd argument expected
                    #  `SupportsRSub[Variable[_T], SupportsAbs[SupportsRound[object]]]`
                    #  but got `Union[bool, float, int]`.
                    so_ind_tnsr[0, 0].item(),
                )

        # Test with signed
        base_adapter = get_adapter(modular=True)

        # adding a categorical feature
        cat_adapter = copy.deepcopy(base_adapter)
        digest = cat_adapter.generator.search_space_digest
        digest.categorical_features = [0]

        sobol_kwargs = {"input_qmc": True, "num_mc_samples": 10}
        seed = 1234
        for adapter in [base_adapter, cat_adapter]:
            discrete_features = (
                adapter.generator.search_space_digest.categorical_features
            )
            with self.subTest(adapter=adapter):
                set_rng_seed(seed)
                # Unsigned
                ind_dict = ax_parameter_sens(
                    adapter=adapter,
                    metrics=None,
                    order="total",
                    signed=False,
                    **sobol_kwargs,
                )
                ind_deriv = compute_derivatives_from_model_list(
                    model_list=[adapter.generator.surrogate.model],
                    bounds=torch.tensor(adapter.generator.search_space_digest.bounds).T,
                    discrete_features=discrete_features,
                    **sobol_kwargs,
                )
                set_rng_seed(seed)  # reset seed to keep discrete features the same
                cat_indices = adapter.generator.search_space_digest.categorical_features
                ind_dict_signed = ax_parameter_sens(
                    adapter=adapter,
                    metrics=None,
                    order="total",
                    signed=True,
                    **sobol_kwargs,
                )
                for i, pname in enumerate(["x1", "x2"]):
                    if i in cat_indices:  # special case for categorical features
                        expected_sign = 1
                    else:
                        expected_sign = torch.sign(ind_deriv[0, i]).item()

                    self.assertEqual(
                        expected_sign,
                        math.copysign(1, ind_dict_signed["branin"][pname]),
                    )
                    self.assertAlmostEqual(
                        (expected_sign * ind_dict["branin"][pname]).item(),
                        ind_dict_signed["branin"][pname],
                    )  # signed
                    self.assertTrue(ind_dict["branin"][pname] >= 0)  # unsigned

    def test_SobolGPSampling(self) -> None:
        bounds = torch.tensor([(0.0, 1.0) for _ in range(2)]).t()
        sensitivity_sampling = SobolSensitivityGPSampling(
            self.model,
            num_mc_samples=10,
            num_gp_samples=10,
            bounds=bounds,
            second_order=True,
        )
        first_order = sensitivity_sampling.first_order_indices()
        total_order = sensitivity_sampling.total_order_indices()
        second_order = sensitivity_sampling.second_order_indices()
        self.assertIsInstance(first_order, Tensor)
        self.assertIsInstance(total_order, Tensor)
        self.assertIsInstance(second_order, Tensor)
        self.assertEqual(first_order.shape, torch.Size([2, 3]))
        self.assertEqual(total_order.shape, torch.Size([2, 3]))
        self.assertEqual(second_order.shape, torch.Size([1, 3]))

        sensitivity_sampling_bootstrap = SobolSensitivityGPSampling(
            self.model,
            num_mc_samples=10,
            num_gp_samples=10,
            bounds=bounds,
            second_order=True,
            num_bootstrap_samples=10,
        )
        first_order = sensitivity_sampling_bootstrap.first_order_indices()
        total_order = sensitivity_sampling_bootstrap.total_order_indices()
        second_order = sensitivity_sampling_bootstrap.second_order_indices()
        self.assertIsInstance(first_order, Tensor)
        self.assertIsInstance(total_order, Tensor)
        self.assertIsInstance(second_order, Tensor)
        self.assertEqual(first_order.shape, torch.Size([2, 5]))
        self.assertEqual(total_order.shape, torch.Size([2, 5]))
        self.assertEqual(second_order.shape, torch.Size([1, 5]))

        with self.assertRaises(ValueError):
            sensitivity_sampling = SobolSensitivityGPSampling(
                self.model,
                num_mc_samples=10,
                num_gp_samples=10,
                bounds=bounds,
                second_order=False,
            )
            first_order = sensitivity_sampling.first_order_indices()
            total_order = sensitivity_sampling.total_order_indices()
            second_order = sensitivity_sampling.second_order_indices()

        discrete_feature = 0
        sensitivity_sampling_discrete = SobolSensitivityGPSampling(
            self.model,
            num_mc_samples=10,
            num_gp_samples=10,
            bounds=bounds,
            second_order=True,
            discrete_features=[discrete_feature],
        )
        sens = sensitivity_sampling_discrete.sensitivity
        A = sens.A
        B = sens.B
        Arnd = A.round()
        Brnd = B.round()
        # testing that the discrete feature is integer valued
        self.assertAllClose(Arnd[:, discrete_feature], A[:, discrete_feature])
        self.assertAllClose(Brnd[:, discrete_feature], B[:, discrete_feature])

        # testing that the other features are not integer valued
        self.assertFalse(torch.allclose(Arnd, A))
        self.assertFalse(torch.allclose(Brnd, B))

    def test_Sobol_raises(self) -> None:
        adapter = get_adapter(modular=True)
        with self.assertRaisesRegex(
            NotImplementedError,
            "Order third and fourth is not supported. Plese choose one of"
            " 'first', 'total' or 'second'.",
        ):
            ax_parameter_sens(
                adapter=adapter,
                metrics=None,
                order="third and fourth",
                signed=False,
            )

    def test_Sobol_computed_correctly_first_order(self) -> None:
        bounds = torch.tensor([[0.0, 0.0], [1.0, 1.0]])
        sensitivity_x1 = SobolSensitivity(
            bounds, input_function=lambda x: x[:, 0], num_mc_samples=10
        )
        sensitivity_x2 = SobolSensitivity(bounds, input_function=lambda x: x[:, 1])

        sensitivity_x1.evalute_function()
        sensitivity_x2.evalute_function()
        fo_indices1 = sensitivity_x1.first_order_indices()
        fo_indices2 = sensitivity_x2.first_order_indices()
        self.assertEqual(fo_indices1[0] / fo_indices1.sum(), 1.0)
        self.assertEqual(fo_indices1[1] / fo_indices1.sum(), 0.0)
        self.assertEqual(fo_indices2[1] / fo_indices2.sum(), 1.0)
        self.assertEqual(fo_indices2[0] / fo_indices2.sum(), 0.0)

    def test_Sobol_computed_correctly_second_order(self) -> None:
        bounds = torch.tensor([[0.0, 0.0], [1.0, 1.0]])
        sensitivity_x1x2 = SobolSensitivity(
            bounds,
            input_function=lambda x: Tensor([1234.5]),  # not used
            second_order=True,
            num_mc_samples=10,
        )
        f_A = torch.rand(10)
        f_B = -f_A
        sensitivity_x1x2.f_total_var = torch.cat((f_A, f_B)).var()
        sensitivity_x1x2.f_A = f_A
        sensitivity_x1x2.f_B = -f_A
        sensitivity_x1x2.f_ABis = [f_A, f_A]
        sensitivity_x1x2.f_BAis = [f_A, -f_B]

        fo_indices = sensitivity_x1x2.first_order_indices()
        so_indices = sensitivity_x1x2.second_order_indices()

        self.assertTrue(torch.all(fo_indices == 0.0))
        self.assertTrue(torch.all(so_indices > 0.0))

    def test_DerivativeGp(self) -> None:
        test_x = torch.rand(2, 2)
        posterior = posterior_derivative(self.model, test_x, kernel_type="matern")
        self.assertIsInstance(posterior, MultivariateNormal)

        with self.assertRaises(ValueError):
            posterior = posterior_derivative(self.model, test_x, kernel_type="xyz")

    def test_sample_discrete_parameters(self) -> None:
        dim = 5
        bounds = torch.stack((torch.zeros(dim), torch.arange(1, dim + 1)))
        num_mc_samples = 8
        A = unnormalize(torch.rand(num_mc_samples, dim), bounds=bounds)
        discrete_features, continuous_features = [1, 3], [0, 2, 4]
        B = sample_discrete_parameters(
            input_mc_samples=A.clone(),
            discrete_features=discrete_features,
            bounds=bounds,
            num_mc_samples=num_mc_samples,
        )
        self.assertTrue(  # Non-discrete parameters should be untouched
            torch.equal(A[:, continuous_features], B[:, continuous_features])
        )
        for i in discrete_features:  # Make sure we sampled integers in the right range
            self.assertTrue(B[:, i].min() >= bounds[0, i])
            self.assertTrue(B[:, i].max() <= bounds[1, i])
            self.assertAllClose(B[:, i], B[:, i].round())
        # discrete_features=None should be a no-op
        B = sample_discrete_parameters(
            input_mc_samples=A.clone(),
            discrete_features=None,
            bounds=bounds,
            num_mc_samples=num_mc_samples,
        )
        self.assertTrue(torch.equal(A, B))



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import functools
import warnings
from collections.abc import Callable, Mapping, Sequence

import numpy as np
import numpy.typing as npt
from ax.core.types import TGenMetadata, TParamValue, TParamValueList
from ax.generators.discrete.ashr_utils import Ashr, GaussianMixture
from ax.generators.discrete.thompson import ThompsonSampler
from ax.generators.discrete_base import DiscreteGenerator
from ax.generators.types import TConfig
from ax.utils.common.docutils import copy_doc
from pyre_extensions import none_throws


class EBAshr(ThompsonSampler):
    """
    Determines feasible arms and sorts them by the objective
    improvements. Feasibile arms are required to have probability
    of feasibility with respect to all metrics in objective and constraints
    above 1-regression_prob_threshold, otherwise they are deemed infeasible.
    These probabilities are computed via the discrete emprirical Bayes Ashr model.
    Shrinkage: Ashr model provides shrinkage to metric outcomes across arms
    with model fitted separately for each metric.
    Prior: Ashr model uses a prior consisting of mixture of Gaussian
    distributions, all of which are centered at zero. The number of Gaussians
    in the prior and their variances are chosen based on observed data to cover
    the whole data range. One of the variances is chosen to be zero so that
    one of the Gaussians in the prior mixture is a point mass at zero.
    The mixture proportions in the prior are learned based on observed outcome
    data (Ys) and their standard errors (Yvars) via standard empirical Bayes
    methodology. The method operates in the outcome space and it doesn't use
    the space of features (Xs).
    """

    def __init__(
        self,
        eb_penalty_param: float = 1.0,
        eb_nsteps: int = 1000,
        eb_threshold: float = 10e-4,
        eb_grid_param: float = 2,
        min_variance_threshold: float = 10e-16,
    ) -> None:
        r"""
        Args:
            eb_penalty_param: Penalty parameter, part of Ashr model log-likelihood,
                represents the amount of shrinkage towards zero. Penalty parameter takes
                real values no smaller than 1. The default 1 means there will be regular
                shrinkage from the Bayes model and the prior but not additional
                shrinkage. Parameter value above one encourages overestimation of the
                null (zero) class so there is additional shrinkage. Recommended
                parameter range is [1, 10].
            eb_nsteps: The number of steps in Ashr fitting EM algorithm.
            eb_threshold: The precision threshold in Ashr fitting EM algorithm.
            eb_grid_param: The grid parameter affecting the number of Gaaussians in the
                prior mixture for Ashr prior distribution.
            min_variance_thresold: All observed test groups/arms with variance below
                this threshold will be considered fixed variables with zero variance.
        """
        self.X: Sequence[Sequence[TParamValue]] | None = None
        self.Ys: Sequence[Sequence[float]] | None = None
        self.Yvars: Sequence[Sequence[float]] | None = None
        self.X_to_Ys_and_Yvars: (
            list[dict[TParamValueList, tuple[float, float]]] | None
        ) = None
        # for each metric, posterior_feasibility gives
        # posterior probabilities of feasibility across all arms
        self.posterior_feasibility: (
            list[Callable[[float, float], npt.NDArray]] | None
        ) = None
        self.eb_penalty_param: float = eb_penalty_param
        self.eb_nsteps: int = eb_nsteps
        self.eb_threshold: float = eb_threshold
        self.eb_grid_param: float = eb_grid_param
        self.min_variance_threshold: float = min_variance_threshold

    @property
    def can_model_in_sample(self) -> bool:
        """Whether this generator can model (e.g. apply shrinkage) on observed
        parameterizations (in this case, it needs to support calling `predict`()
        on points in the training data / provided during `fit()`)."""
        return True

    def _fit_Ys_and_Yvars(
        self,
        Ys: Sequence[Sequence[float]],
        Yvars: Sequence[Sequence[float]],
        outcome_names: Sequence[str],
    ) -> tuple[list[list[float]], list[list[float]]]:
        r"""
        Args:
            Ys: List of length m of measured metrics across k arms. The outcomes
                are assumed to be relativized with respect to status quo.
            Yvars: List of length m of measured metrics variances across arms.
            outcome_names: List of m metric names.

        Returns: A tuple containing
            - shrinken outcome estimates and
            - their variances.
        """
        newYs = []
        newYvars = []
        self.posterior_feasibility = []

        for Y, Yvar in zip(Ys, Yvars, strict=True):
            newY: npt.NDArray = np.array(Y, dtype=float)
            newYvar = np.array(Yvar, dtype=float)

            # excluding arms with zero variance, e.g. status quo
            nonconstant_rvs = np.abs(newYvar) > self.min_variance_threshold
            # Case where the standard deviations are not infinitesimal, so we do
            # shrinkage.
            if nonconstant_rvs.any():
                # Ashr model applied to arms with non-zero variance
                stats = newY[nonconstant_rvs]
                variances = newYvar[nonconstant_rvs]

                # setting up the Ashr model
                model = Ashr(Y=stats, Yvar=variances, eb_grid_param=self.eb_grid_param)
                k = len(model.prior_vars)
                # run Ashr fitting procedure
                model_fit = model.fit(
                    lambdas=np.array([self.eb_penalty_param] + [1.0] * (k - 1)),
                    nsteps=self.eb_nsteps,
                    threshold=self.eb_threshold,
                )

                # Ashr model posterior
                posterior = model.posterior(model_fit["weights"])
                newY[nonconstant_rvs] = posterior.means
                newYvar[nonconstant_rvs] = posterior.vars

                f_posterior_feas: Callable[[float, float], npt.NDArray] = (
                    functools.partial(
                        posterior_feasibility_util,
                        posterior=posterior,
                        nonconstant_rvs=nonconstant_rvs,
                        Y=newY,
                    )
                )
            else:
                f_posterior_feas: Callable[[float, float], npt.NDArray] = (
                    functools.partial(no_model_feasibility_util, Y=newY)
                )

            none_throws(self.posterior_feasibility).append(f_posterior_feas)

            newYs.append(list(newY))
            newYvars.append(list(newYvar))

        return newYs, newYvars

    def _check_metric_direction(
        self,
        objective_weights: npt.NDArray,
        outcome_constraints: tuple[npt.NDArray, npt.NDArray] | None = None,
    ) -> npt.NDArray:
        r"""
        Args:
            objective_weights: A length m array corresponding to the metrics weights in
                a maximization problem. A positive weight means the corresponding metric
                succeeding in the positive direction. A negative weight means
                the corresponding metric succeeding in the negative direction.
                Outcomes that are modeled but not part of the objective get objective
                weight 0.
            outcome_constraints: A tuple of (A, b), where A is of size
                (num of constraints x m), m is the number of outputs/metrics,
                and b is of size (num of constraints x 1).

        Returns:
            A boolean array of length m indicating the direction of success
            for each metric.
        """
        # metrics in the objective
        upper_is_better = objective_weights > 0

        # constraints
        if outcome_constraints:
            A, b = outcome_constraints
            upper_is_better_constraints = -np.apply_along_axis(
                np.sum, 0, A
            )  # column-wise (arm-wise)
            objectives = objective_weights != 0
            upper_is_better[~objectives] = upper_is_better_constraints[~objectives]

        return upper_is_better

    def _get_regression_indicator(
        self,
        objective_weights: npt.NDArray,
        outcome_constraints: tuple[npt.NDArray, npt.NDArray] | None = None,
        regression_prob_threshold: float = 0.90,
    ) -> tuple[npt.NDArray, npt.NDArray]:
        r"""
        Args:
            regression_prob_threshold: Arms having a regression probability with respect
                to any metric (in either objective or constraints) above a given
                threshold are marked as regressions.

        Returns: A tuple containing:
            - A length k array containing regression indicators for each of the k arms.
            - (num of arms, num of metrics) matrix of arm, metric pairs, (i, j)-entry
            corresponds to the probability of infeasibility or i-th arm and j-th metric.
        """
        num_metrics = len(none_throws(self.Ys))
        num_arms = len(none_throws(self.Ys)[0])
        prob_infeasibility = np.zeros(
            (num_arms, num_metrics), dtype=float
        )  # (num of arms) x (num of metrics)

        upper_is_better = self._check_metric_direction(
            objective_weights=objective_weights, outcome_constraints=outcome_constraints
        )
        if np.sum(abs(objective_weights) > 0) > 1:
            warnings.warn(
                "In case of multi-objective adding metric values together might"
                " not lead to a meaningful result.",
                stacklevel=2,
            )
        # compute probabilities of feasibility for metrics in the objective
        # equals regression probabilities in this case
        for i in range(num_metrics):
            if objective_weights[i] != 0.0:
                lb = 0.0 if upper_is_better[i] else -np.inf
                ub = np.inf if upper_is_better[i] else 0.0
                prob_infeasibility[:, i] = 1.0 - none_throws(
                    self.posterior_feasibility
                )[i](lb, ub)  # per i-th metric, across arms

        # compute probabilities of feasibility for constraints
        if outcome_constraints:
            A, b = outcome_constraints

            if np.any(
                np.apply_along_axis(lambda x: np.linalg.norm(x, ord=0) > 1, 1, A)
            ):
                raise ValueError(
                    "Only one metric per constraint allowed. Scalarized "
                    " OutcomeConstraint with multiple metrics per constraint not "
                    "supported."
                )

            for i in range(A.shape[1]):
                if np.linalg.norm(A[:, i], ord=1) == 0:
                    continue
                upper = A[:, i] > 0
                lower = A[:, i] < 0
                ub = np.min(b[upper] / A[upper, i]) if np.any(upper) else np.inf
                lb = np.max(b[lower] / A[lower, i]) if np.any(lower) else -np.inf
                # per i-th metric, across arms; probability a metric strictly
                # smaller than lb or strictly larger than ub
                prob_infeasibility[:, i] = 1.0 - none_throws(
                    self.posterior_feasibility
                )[i](lb, ub)

        # for each arm check if it is infeasible with respect to any metric
        regressions = np.apply_along_axis(
            np.any, 1, prob_infeasibility >= regression_prob_threshold
        )

        return regressions, prob_infeasibility

    def _get_success_measurement(self, objective_weights: npt.NDArray) -> npt.NDArray:
        r"""
        Returns:
            A length k array returning a measure of success for each of k arms.
        """

        posterior_means = np.array(none_throws(self.Ys)).T
        success = np.dot(posterior_means, objective_weights)
        return success

    @copy_doc(DiscreteGenerator.gen)
    def gen(
        self,
        n: int,
        parameter_values: Sequence[Sequence[TParamValue]],
        objective_weights: npt.NDArray | None,
        outcome_constraints: tuple[npt.NDArray, npt.NDArray] | None = None,
        fixed_features: Mapping[int, TParamValue] | None = None,
        pending_observations: Sequence[Sequence[Sequence[TParamValue]]] | None = None,
        model_gen_options: TConfig | None = None,
    ) -> tuple[list[Sequence[TParamValue]], list[float], TGenMetadata]:
        if objective_weights is None:
            raise ValueError("EB Ashr model requires objective weights.")

        model_gen_options = model_gen_options or {}
        regression_prob_threshold = model_gen_options.get(
            "regression_prob_threshold", 0.90
        )
        if not isinstance(regression_prob_threshold, float):
            raise TypeError(
                "`regression_prob_threshold` is required among `model_gen_kwargs` \
                and must be set to a float."
            )

        # prob_infeasibility: matrix containing probabilities of infeasibilities
        # for each arm, metric pair
        regression, prob_infeasibility = self._get_regression_indicator(
            objective_weights=objective_weights,
            outcome_constraints=outcome_constraints,
            regression_prob_threshold=regression_prob_threshold,
        )
        success = self._get_success_measurement(objective_weights=objective_weights)

        arms = none_throws(self.X)
        sorted_arms = sorted(
            zip(arms, list(success), list(regression), np.arange(len(arms))),
            key=lambda x: x[1] if not x[2] else -1.0,
            reverse=True,
        )

        top_arms = [sorted_arms[i][0] for i in range(n)]

        return (
            top_arms,
            [1.0] * len(top_arms),  # uniform weights
            {
                "regression": regression,
                "success": success,
                "prob_infeasibility": prob_infeasibility,
                "best_x": sorted_arms[0][0],
            },
        )


def posterior_feasibility_util(
    lb: float,
    ub: float,
    posterior: GaussianMixture,
    nonconstant_rvs: npt.NDArray,
    Y: npt.NDArray,
) -> npt.NDArray:
    r"""
    Probabilities of a posterior rv being inside the given interval [lb, up]
    with bounds included per single metric across arms.

    Args:
        lb: Lower bound on the random variable.
        ub: Upper bound on the random variable.
        posterior: Posterior classs computing posterior distribution for all
            non-degenerate (non-constant) arms.
        nonconstant_rvs: Length k boolean vector indicating which variables
            are constants.
        Y: Length k outcome vector.

    Returns:
        Length k array of probabilities of feasibility across k arms.
    """
    # Pr(m>ub)
    # TODO: jmarkovic simplify this directly in Ashr
    upper_tail = (
        posterior.tail_probabilities(left_tail=False, threshold=ub)
        + posterior.weights[:, 0] * (ub < 0)
        if ub < np.inf
        else 0
    )
    # Pr(m<lb)
    lower_tail = (
        posterior.tail_probabilities(left_tail=True, threshold=lb)
        + posterior.weights[:, 0] * (lb > 0)
        if lb > -np.inf
        else 0
    )

    prob_feas = np.zeros_like(nonconstant_rvs, dtype=float)
    # pyre-fixme[58]: `-` is not supported for operand types `float` and
    #  `Union[np.ndarray[typing.Any, np.dtype[typing.Any]], int]`.
    prob_feas[nonconstant_rvs] = 1.0 - upper_tail - lower_tail
    prob_feas[~nonconstant_rvs] = (Y[~nonconstant_rvs] >= lb) & (
        Y[~nonconstant_rvs] <= ub
    )
    return prob_feas


def no_model_feasibility_util(lb: float, ub: float, Y: npt.NDArray) -> npt.NDArray:
    r"""
    Probabilities (0 or 1) of whether Y is inside [lb, ub].

    This covers the degenerate case when all arms have zero variance, and
    `posterior_feasibility_util` is not usable because there is no posterior.
    """
    return (Y >= lb) & (Y <= ub)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from unittest.mock import patch

import torch
from ax.benchmark.benchmark_problem import BenchmarkProblem
from ax.benchmark.benchmark_test_functions.botorch_test import BoTorchTestFunction

from ax.benchmark.problems.synthetic.discretized.mixed_integer import (
    get_discrete_ackley,
    get_discrete_hartmann,
    get_discrete_rosenbrock,
)
from ax.core.parameter import ParameterType
from ax.utils.common.testutils import TestCase
from botorch.test_functions.synthetic import Ackley, Hartmann, Rosenbrock
from pyre_extensions import assert_is_instance


class MixedIntegerProblemsTest(TestCase):
    def test_problems(self) -> None:
        for problem_cls, constructor, dim, dim_int in (
            (Hartmann, get_discrete_hartmann, 6, 4),
            (Ackley, get_discrete_ackley, 13, 10),
            (Rosenbrock, get_discrete_rosenbrock, 10, 6),
        ):
            name = problem_cls.__name__
            problem = constructor()
            self.assertEqual(f"Discrete {name}", problem.name)
            test_function = assert_is_instance(
                problem.test_function, BoTorchTestFunction
            )
            botorch_problem = test_function.botorch_problem
            self.assertIsInstance(botorch_problem, problem_cls)
            self.assertEqual(len(problem.search_space.parameters), dim)
            self.assertEqual(
                sum(
                    p.parameter_type == ParameterType.INT
                    for p in problem.search_space.parameters.values()
                ),
                dim_int,
            )
            # Check that the underlying problem has the correct bounds.
            if name == "Rosenbrock":
                expected_bounds = [(-5.0, 10.0) for _ in range(dim)]
            else:
                expected_bounds = [(0.0, 1.0) for _ in range(dim)]
            self.assertEqual(botorch_problem._bounds, expected_bounds)
            self.assertGreaterEqual(problem.optimal_value, problem_cls().optimal_value)

        # Test that they match correctly to the original problems.
        cases: list[tuple[BenchmarkProblem, dict[str, float], torch.Tensor]] = [
            (
                get_discrete_hartmann(),
                {f"x{i + 1}": 0.0 for i in range(6)},
                torch.zeros(6, dtype=torch.double),
            ),
            (
                get_discrete_hartmann(),
                {"x1": 3, "x2": 3, "x3": 19, "x4": 19, "x5": 1.0, "x6": 1.0},
                torch.ones(6, dtype=torch.double),
            ),
            (
                get_discrete_ackley(),
                {f"x{i + 1}": 0.0 for i in range(13)},
                torch.zeros(13, dtype=torch.double),
            ),
            (
                get_discrete_ackley(),
                {
                    **{f"x{i + 1}": 2 for i in range(0, 5)},
                    **{f"x{i + 1}": 4 for i in range(5, 10)},
                    **{f"x{i + 1}": 1.0 for i in range(10, 13)},
                },
                torch.ones(13, dtype=torch.double),
            ),
            (
                get_discrete_rosenbrock(),
                {f"x{i + 1}": 0.0 for i in range(10)},
                torch.full((10,), -5.0, dtype=torch.double),
            ),
            (
                get_discrete_rosenbrock(),
                {
                    **{f"x{i + 1}": 3 for i in range(0, 6)},
                    **{f"x{i + 1}": 1.0 for i in range(6, 10)},
                },
                torch.full((10,), 10.0, dtype=torch.double),
            ),
        ]

        for problem, params, expected_arg in cases:
            test_function = assert_is_instance(
                problem.test_function, BoTorchTestFunction
            )
            with patch.object(
                test_function.botorch_problem,
                attribute="evaluate_true",
                wraps=test_function.botorch_problem.evaluate_true,
            ) as mock_call:
                test_function.evaluate_true(params=params)
            actual = mock_call.call_args.kwargs["X"]
            self.assertAllClose(actual, expected_arg)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


from ax.adapter.base import DataLoaderConfig
from ax.adapter.discrete import DiscreteAdapter
from ax.adapter.random import RandomAdapter
from ax.adapter.registry import (
    _extract_model_state_after_gen,
    Cont_X_trans,
    Generators,
    MODEL_KEY_TO_MODEL_SETUP,
    Models,
    Y_trans,
)
from ax.adapter.torch import TorchAdapter
from ax.core.observation import ObservationFeatures
from ax.core.optimization_config import MultiObjectiveOptimizationConfig
from ax.exceptions.core import UserInputError
from ax.generators.discrete.eb_thompson import EmpiricalBayesThompsonSampler
from ax.generators.discrete.thompson import ThompsonSampler
from ax.generators.random.sobol import SobolGenerator
from ax.generators.torch.botorch_modular.acquisition import Acquisition
from ax.generators.torch.botorch_modular.generator import BoTorchGenerator
from ax.generators.torch.botorch_modular.kernels import ScaleMaternKernel
from ax.generators.torch.botorch_modular.surrogate import (
    ModelConfig,
    Surrogate,
    SurrogateSpec,
)
from ax.utils.common.kwargs import get_function_argument_names
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import (
    get_branin_experiment,
    get_branin_experiment_with_status_quo_trials,
    get_branin_optimization_config,
    get_branin_search_space,
    get_factorial_experiment,
)
from ax.utils.testing.mock import mock_botorch_optimize
from botorch.acquisition.monte_carlo import qExpectedImprovement
from botorch.models.fully_bayesian import SaasFullyBayesianSingleTaskGP
from botorch.models.fully_bayesian_multitask import SaasFullyBayesianMultiTaskGP
from botorch.models.gp_regression import SingleTaskGP
from botorch.models.model_list_gp_regression import ModelListGP
from botorch.models.multitask import MultiTaskGP
from botorch.utils.types import DEFAULT
from gpytorch.kernels.matern_kernel import MaternKernel
from gpytorch.kernels.rbf_kernel import RBFKernel
from gpytorch.kernels.scale_kernel import ScaleKernel
from gpytorch.mlls.exact_marginal_log_likelihood import ExactMarginalLogLikelihood
from gpytorch.priors.torch_priors import GammaPrior, LogNormalPrior
from pyre_extensions import assert_is_instance


class ModelRegistryTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.maxDiff = None

    @mock_botorch_optimize
    def test_botorch_modular(self) -> None:
        exp = get_branin_experiment(with_batch=True)
        exp.trials[0].run()
        gpei = Generators.BOTORCH_MODULAR(
            # Model kwargs
            acquisition_class=Acquisition,
            botorch_acqf_class=qExpectedImprovement,
            botorch_acqf_options={"best_f": 0.0},
            # Adapter kwargs
            experiment=exp,
            data=exp.fetch_data(),
        )
        self.assertIsInstance(gpei, TorchAdapter)
        generator = assert_is_instance(gpei.generator, BoTorchGenerator)
        self.assertEqual(generator.botorch_acqf_class, qExpectedImprovement)
        self.assertEqual(generator.acquisition_class, Acquisition)
        self.assertEqual(generator._botorch_acqf_options, {"best_f": 0.0})
        self.assertIsInstance(generator.surrogate, Surrogate)
        # SingleTaskGP should be picked.
        self.assertIsInstance(generator.surrogate.model, SingleTaskGP)

        gr = gpei.gen(n=1)
        self.assertIsNotNone(gr.best_arm_predictions)

    @mock_botorch_optimize
    def test_SAASBO(self) -> None:
        exp = get_branin_experiment()
        sobol = Generators.SOBOL(experiment=exp)
        self.assertIsInstance(sobol, RandomAdapter)
        for _ in range(5):
            sobol_run = sobol.gen(n=1)
            self.assertEqual(sobol_run._model_key, "Sobol")
            exp.new_batch_trial().add_generator_run(sobol_run).run()
        saasbo = Generators.SAASBO(experiment=exp, data=exp.fetch_data())
        self.assertIsInstance(saasbo, TorchAdapter)
        self.assertEqual(saasbo._model_key, "SAASBO")
        generator = assert_is_instance(saasbo.generator, BoTorchGenerator)
        surrogate_spec = generator.surrogate_spec
        self.assertEqual(
            surrogate_spec,
            SurrogateSpec(
                model_configs=[
                    ModelConfig(
                        botorch_model_class=SaasFullyBayesianSingleTaskGP, name="SAASBO"
                    )
                ]
            ),
        )
        self.assertEqual(
            generator.surrogate.surrogate_spec.model_configs[0].botorch_model_class,
            SaasFullyBayesianSingleTaskGP,
        )

    @mock_botorch_optimize
    def test_enum_sobol_legacy_GPEI(self) -> None:
        """Tests Sobol and Legacy GPEI instantiation through the Generators enum."""
        exp = get_branin_experiment()
        # Check that factory generates a valid sobol adapter.
        sobol = Generators.SOBOL(experiment=exp)
        self.assertIsInstance(sobol, RandomAdapter)
        for _ in range(5):
            sobol_run = sobol.gen(n=1)
            self.assertEqual(sobol_run._model_key, "Sobol")
            exp.new_batch_trial().add_generator_run(sobol_run).run()
        # Check that factory generates a valid GP+EI adapter.
        exp.optimization_config = get_branin_optimization_config()
        gpei = Generators.LEGACY_BOTORCH(experiment=exp, data=exp.fetch_data())
        self.assertIsInstance(gpei, TorchAdapter)
        self.assertEqual(gpei._model_key, "Legacy_GPEI")
        botorch_defaults = "ax.generators.torch.botorch_defaults"
        # Check that the callable kwargs and the torch kwargs were recorded.
        self.assertEqual(
            gpei._model_kwargs,
            {
                "acqf_constructor": {
                    "is_callable_as_path": True,
                    "value": f"{botorch_defaults}.get_qLogNEI",
                },
                "acqf_optimizer": {
                    "is_callable_as_path": True,
                    "value": f"{botorch_defaults}.scipy_optimizer",
                },
                "model_constructor": {
                    "is_callable_as_path": True,
                    "value": f"{botorch_defaults}.get_and_fit_model",
                },
                "model_predictor": {
                    "is_callable_as_path": True,
                    "value": "ax.generators.torch.utils.predict_from_model",
                },
                "best_point_recommender": {
                    "is_callable_as_path": True,
                    "value": f"{botorch_defaults}.recommend_best_observed_point",
                },
                "refit_on_cv": False,
                "warm_start_refitting": True,
                "use_input_warping": False,
                "use_loocv_pseudo_likelihood": False,
                "prior": None,
            },
        )
        self.assertEqual(
            gpei._bridge_kwargs,
            {
                "transform_configs": None,
                "torch_device": None,
                "optimization_config": None,
                "transforms": Cont_X_trans + Y_trans,
                "expand_model_space": True,
                "fit_tracking_metrics": True,
                "fit_on_init": True,
                "default_model_gen_options": None,
                "data_loader_config": DataLoaderConfig(
                    fit_only_completed_map_metrics=True,
                ),
                # now passed through the data loader config
                "fit_only_completed_map_metrics": None,
                "fit_out_of_design": None,
                "fit_abandoned": None,
            },
        )
        self.assertIsInstance(gpei._data_loader_config, DataLoaderConfig)
        self.assertEqual(gpei._data_loader_config.fit_only_completed_map_metrics, True)
        self.assertEqual(gpei._data_loader_config.fit_out_of_design, False)
        self.assertEqual(gpei._data_loader_config.fit_abandoned, False)

        prior_kwargs = {"lengthscale_prior": GammaPrior(6.0, 6.0)}
        gpei = Generators.LEGACY_BOTORCH(
            experiment=exp,
            data=exp.fetch_data(),
            search_space=exp.search_space,
            prior=prior_kwargs,
        )
        self.assertIsInstance(gpei, TorchAdapter)
        self.assertEqual(
            gpei._model_kwargs["prior"],  # pyre-ignore
            prior_kwargs,
        )

    def test_enum_model_kwargs(self) -> None:
        """Tests that kwargs are passed correctly when instantiating through the
        Generators enum."""
        exp = get_branin_experiment()
        sobol = Generators.SOBOL(
            experiment=exp, init_position=2, scramble=False, seed=239
        )
        self.assertIsInstance(sobol, RandomAdapter)
        for _ in range(5):
            sobol_run = sobol.gen(1)
            exp.new_batch_trial().add_generator_run(sobol_run).run()

    def test_enum_factorial(self) -> None:
        """Tests factorial instantiation through the Generators enum."""
        exp = get_factorial_experiment()
        factorial = Generators.FACTORIAL(experiment=exp)
        self.assertIsInstance(factorial, DiscreteAdapter)
        factorial_run = factorial.gen(n=-1)
        self.assertEqual(len(factorial_run.arms), 24)

    def test_enum_empirical_bayes_thompson(self) -> None:
        """Tests EB/TS instantiation through the Generators enum."""
        exp = get_factorial_experiment()
        factorial = Generators.FACTORIAL(experiment=exp)
        self.assertIsInstance(factorial, DiscreteAdapter)
        factorial_run = factorial.gen(n=-1)
        exp.new_batch_trial().add_generator_run(factorial_run).run().mark_completed()
        data = exp.fetch_data()
        eb_thompson = Generators.EMPIRICAL_BAYES_THOMPSON(
            experiment=exp, data=data, min_weight=0.0
        )
        self.assertIsInstance(eb_thompson, DiscreteAdapter)
        self.assertIsInstance(eb_thompson.generator, EmpiricalBayesThompsonSampler)
        thompson_run = eb_thompson.gen(n=5)
        self.assertEqual(len(thompson_run.arms), 5)

    def test_enum_thompson(self) -> None:
        """Tests TS instantiation through the Generators enum."""
        exp = get_factorial_experiment()
        factorial = Generators.FACTORIAL(experiment=exp)
        self.assertIsInstance(factorial, DiscreteAdapter)
        factorial_run = factorial.gen(n=-1)
        exp.new_batch_trial().add_generator_run(factorial_run).run().mark_completed()
        data = exp.fetch_data()
        thompson = Generators.THOMPSON(experiment=exp, data=data)
        self.assertIsInstance(thompson.generator, ThompsonSampler)

    def test_enum_uniform(self) -> None:
        """Tests uniform random instantiation through the Generators enum."""
        exp = get_branin_experiment()
        uniform = Generators.UNIFORM(experiment=exp)
        self.assertIsInstance(uniform, RandomAdapter)
        uniform_run = uniform.gen(n=5)
        self.assertEqual(len(uniform_run.arms), 5)

    def test_view_defaults(self) -> None:
        """Checks that kwargs are correctly constructed from default kwargs +
        standard kwargs."""
        self.assertEqual(
            Generators.SOBOL.view_defaults(),
            (
                {
                    "seed": None,
                    "deduplicate": True,
                    "init_position": 0,
                    "scramble": True,
                    "generated_points": None,
                    "fallback_to_sample_polytope": False,
                },
                {
                    "optimization_config": None,
                    "transforms": Cont_X_trans,
                    "transform_configs": None,
                    "fit_out_of_design": None,  # False by DataLoaderConfig default
                    "fit_abandoned": None,  # False by DataLoaderConfig default
                    "data_loader_config": None,
                    "fit_tracking_metrics": True,
                    "fit_on_init": True,
                },
            ),
        )
        self.assertTrue(
            all(
                kw in Generators.SOBOL.view_kwargs()[0]
                for kw in ["seed", "deduplicate", "init_position", "scramble"]
            ),
            all(
                kw in Generators.SOBOL.view_kwargs()[1]
                for kw in [
                    "search_space",
                    "model",
                    "transforms",
                    "experiment",
                    "data",
                    "transform_configs",
                    "expand_model_space",
                    "fit_out_of_design",
                    "fit_abandoned",
                    "fit_tracking_metrics",
                    "fit_on_init",
                ]
            ),
        )

    def test_ModelSetups_do_not_share_kwargs(self) -> None:
        """Tests that none of the preset model and adapter combinations share a
        kwarg.
        """
        for model_setup_info in MODEL_KEY_TO_MODEL_SETUP.values():
            model_class = model_setup_info.model_class
            adapter_class = model_setup_info.adapter_class
            model_args = set(get_function_argument_names(model_class))
            bridge_args = set(get_function_argument_names(adapter_class))
            # Intersection of two sets should be empty
            self.assertEqual(model_args & bridge_args, set())

    @mock_botorch_optimize
    def test_ST_MTGP(self, use_saas: bool = False) -> None:
        """Tests single type MTGP via Modular BoTorch instantiation
        with both single & multi objective optimization."""
        for exp in [
            get_branin_experiment_with_status_quo_trials(num_sobol_trials=2),
            get_branin_experiment_with_status_quo_trials(
                num_sobol_trials=2, multi_objective=True
            ),
        ]:
            # testing custom and default kernel for a surrogate
            surrogates = (
                [None]
                if use_saas
                else [
                    Surrogate(
                        surrogate_spec=SurrogateSpec(
                            model_configs=[
                                ModelConfig(
                                    botorch_model_class=MultiTaskGP,
                                    mll_class=ExactMarginalLogLikelihood,
                                    covar_module_class=ScaleMaternKernel,
                                    covar_module_options={
                                        "ard_num_dims": DEFAULT,
                                        "lengthscale_prior": GammaPrior(6.0, 3.0),
                                        "outputscale_prior": GammaPrior(2.0, 0.15),
                                        "batch_shape": DEFAULT,
                                    },
                                    model_options={},
                                )
                            ],
                            allow_batched_models=False,
                        )
                    ),
                    None,
                ]
            )

            for surrogate, default_model in zip(surrogates, (False, True)):
                constructor = Generators.SAAS_MTGP if use_saas else Generators.ST_MTGP
                mtgp = constructor(
                    experiment=exp,
                    data=exp.fetch_data(),
                    surrogate=surrogate,
                )
                self.assertIsInstance(mtgp, TorchAdapter)
                generator = assert_is_instance(mtgp.generator, BoTorchGenerator)
                self.assertEqual(generator.acquisition_class, Acquisition)
                is_moo = isinstance(
                    exp.optimization_config, MultiObjectiveOptimizationConfig
                )
                if is_moo:
                    self.assertIsInstance(generator.surrogate.model, ModelListGP)
                    models = generator.surrogate.model.models
                else:
                    models = [generator.surrogate.model]

                for model in models:
                    self.assertIsInstance(
                        model,
                        SaasFullyBayesianMultiTaskGP if use_saas else MultiTaskGP,
                    )
                    data_covar_module, task_covar_module = model.covar_module.kernels
                    if use_saas is False and default_model is False:
                        self.assertIsInstance(data_covar_module, ScaleKernel)
                        base_kernel = data_covar_module.base_kernel
                        self.assertIsInstance(base_kernel, MaternKernel)
                        self.assertEqual(
                            base_kernel.lengthscale_prior.concentration, 6.0
                        )
                        self.assertEqual(base_kernel.lengthscale_prior.rate, 3.0)
                    elif use_saas is False:
                        self.assertIsInstance(data_covar_module, RBFKernel)
                        self.assertIsInstance(
                            data_covar_module.lengthscale_prior, LogNormalPrior
                        )

                gr = mtgp.gen(
                    n=1,
                    fixed_features=ObservationFeatures({}, trial_index=1),
                )
                self.assertEqual(len(gr.arms), 1)

    def test_SAAS_MTGP(self) -> None:
        self.test_ST_MTGP(use_saas=True)

    def test_extract_model_state_after_gen(self) -> None:
        # Test with actual state.
        exp = get_branin_experiment()
        sobol = Generators.SOBOL(experiment=exp)
        gr = sobol.gen(n=1)
        expected_state = sobol.generator._get_state()
        self.assertEqual(gr._model_state_after_gen, expected_state)
        extracted = _extract_model_state_after_gen(
            generator_run=gr, model_class=SobolGenerator
        )
        self.assertEqual(extracted, expected_state)
        # Test with empty state.
        gr._model_state_after_gen = None
        extracted = _extract_model_state_after_gen(
            generator_run=gr, model_class=SobolGenerator
        )
        self.assertEqual(extracted, {})

    def test_deprecation_warning(self) -> None:
        """Tests deprecation warning"""
        with self.assertRaisesRegex(
            DeprecationWarning,
            r"Models is deprecated, use \`ax.adapter.registry.Generators\`"
            r" instead.",
        ):
            Models.BOTORCH_MODULAR

    def test_initialize_from_search_space(self) -> None:
        search_space = get_branin_search_space()
        with self.assertWarnsRegex(
            DeprecationWarning, "Passing in a `search_space` to initialize"
        ):
            adapter = Generators.SOBOL(search_space=search_space)
        self.assertEqual(adapter._model_space, search_space)
        self.assertIsNotNone(adapter._experiment)
        with self.assertRaisesRegex(
            UserInputError,
            "`experiment` is required to initialize a model from registry.",
        ):
            Generators.BOTORCH_MODULAR(search_space=search_space)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from collections import defaultdict

import numpy as np
import numpy.typing as npt
from ax.adapter.discrete import DiscreteAdapter
from ax.adapter.registry import rel_EB_ashr_trans

from ax.core.data import Data
from ax.core.experiment import Experiment
from ax.core.observation import observations_from_data

from ax.exceptions.core import DataRequiredError, UserInputError
from ax.generators.discrete.eb_ashr import EBAshr
from pyre_extensions import assert_is_instance


def detect_regressions_by_trial(
    experiment: Experiment,
    thresholds: dict[str, tuple[float, float]],
    data: Data | None = None,
) -> dict[int, dict[str, dict[str, float]]]:
    r"""
    Identifies all regressing arms across trial along with the metrics they regress.

    Args:
        experiment: Ax experiment.
        thresholds: A dictionary mapping metric names a tuple of
            (threshold size, threshold probability).
        data: Experiment data. If None, use the attached experiment data.

    Returns: A dictionary mapping trial indices to
        dictionaries containing regressing arm names as keys and a dictionary
        with the corresponding regressed metrics and a tuple of regression sizes and
        corresponding regression probabilities as values.

    """
    data = data if data is not None else experiment.lookup_data()

    regressing_arms_metrics_by_trial = {}

    for trial_index, trial_df in data.df.groupby("trial_index"):
        regressing_arms_metrics_by_trial[trial_index] = detect_regressions_single_trial(
            experiment=experiment,
            data=Data(df=trial_df),
            thresholds=thresholds,
        )

    return regressing_arms_metrics_by_trial


def compute_regression_probabilities_single_trial(
    experiment: Experiment,
    size_thresholds: dict[str, float],
    data: Data | None = None,
) -> tuple[list[str | None], list[str], npt.NDArray]:
    r"""
    Computes the probabilities of regression for all arm metric pairs in a single trial.
    If a metric has lower_is_better indicator set to True
    (regresses in positive direction), regression probability is defined as
    Pr(metric value>=(positive) threshold).
    If a metric has lower_is_better indicator set to False
    (regresses in negative direction), regression probability is defined as
    Pr(metric value<=(negative) threshold).

    Args:
        experiment: Ax experiment.
        size_thresholds: A dictionary mapping metric names to threshold sizes.
        data: Experiment data. If None, use the attached experiment data.

    Returns: A tuple containing
        - arm names,
        - metric names, and corresponding
        - an array of size n x m where n is the number of arms and m is
            the number of metrics. Each entry contains the probability of
            regression for that arm-metric pair.

    """
    data = data if data is not None else experiment.lookup_data()

    if data.df.empty:
        raise DataRequiredError("Data must be non-empty.")

    if len(set(data.df["trial_index"])) > 1:
        raise UserInputError("The input data should contain only one trial.")

    # metric names available in the data and also in the given size thresholds dict
    metric_names = set(size_thresholds.keys()).intersection(set(data.df["metric_name"]))

    if len(metric_names) == 0:
        raise ValueError(
            "No common metrics between the provided data and the size thresholds."
            "Need to provide both data and the size thresholds for metrics of interest."
        )

    target_data = Data(df=data.df[data.df["metric_name"].isin(metric_names)])

    adapter = DiscreteAdapter(
        experiment=experiment,
        data=target_data,
        generator=EBAshr(),
        transforms=rel_EB_ashr_trans,
    )

    metric_names = adapter.outcomes

    lower_is_better_indicators = np.array(
        [-1 if experiment.metrics[m].lower_is_better else 1 for m in metric_names]
    )

    # if lower is better: outcome constraints: metric <= (positive) threshold
    # if upper is better: outcome constraints: metric >= (negative) threshold
    A = -np.diag(lower_is_better_indicators)
    b = np.array([abs(size_thresholds[metric]) for metric in metric_names])

    _, regression_probabilities = assert_is_instance(
        adapter.generator, EBAshr
    )._get_regression_indicator(
        objective_weights=np.zeros(len(metric_names)), outcome_constraints=(A, b)
    )

    observations = observations_from_data(
        experiment=experiment,
        data=target_data,
    )

    arm_names = [obs.arm_name for obs in observations]

    return arm_names, metric_names, regression_probabilities


def detect_regressions_single_trial(
    experiment: Experiment,
    thresholds: dict[str, tuple[float, float]],
    data: Data | None = None,
) -> dict[str | None, dict[str, float]]:
    r"""
    Identifies all regressing arms along with the metrics they regress for
    a single trial.

    Args:
        experiment: Ax experiment.
        thresholds: A dictionary mapping metric names a tuple of
            (threshold size, threshold probability).

    Returns: A dictionary with regressing arms as keys and another dictionary of
        corresponding regressed metrics and their corresponding regression probabilities
        as values.

    """
    data = data if data is not None else experiment.lookup_data()

    if len(set(data.df["trial_index"])) > 1:
        raise UserInputError("The input data should contain only one trial.")

    # metric names available in the data and also in the given thresholds
    metric_names = set(thresholds.keys()).intersection(set(data.df["metric_name"]))

    if len(metric_names) == 0:
        raise ValueError(
            "No common metrics between the provided data and the thresholds."
            "Need to provide both data and the size thresholds for metrics of interest."
        )

    size_thresholds = {metric: thresholds[metric][0] for metric in metric_names}

    (
        arm_names,
        metric_names,
        regression_probabilities,
    ) = compute_regression_probabilities_single_trial(
        experiment=experiment,
        data=Data(df=data.df[data.df["metric_name"].isin(metric_names)]),
        size_thresholds=size_thresholds,
    )

    # Regressing arms along with the metrics they regress
    regressing_arms_metrics: dict[str | None, dict[str, float]] = defaultdict(dict)

    for i, arm_name in enumerate(arm_names):
        for j, metric_name in enumerate(metric_names):
            if regression_probabilities[i, j] >= thresholds[metric_name][1]:
                regressing_arms_metrics[arm_name].update(
                    {metric_name: regression_probabilities[i, j]}
                )

    return regressing_arms_metrics



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


import numpy as np
import pandas as pd
import scipy

from ax.core.data import Data
from ax.utils.stats.statstools import relativize
from scipy.stats import norm


def check_experiment_effects_per_metric(
    data: Data,
    objective_names: set[str] | None = None,
    no_effect_alpha: float = 0.05,
) -> pd.DataFrame:
    """Checks whether an experiment has effects per metric separately.
    The results will include the p-value and whether there was an effect
    for each of the available metrics in the data. `check_experiment_effects`
    below checks whether there is any effect overall across all metrics.

    This checks whether a randomization test can show that there are any
    effects whatsoever. This test is performed independently on each metric
    and it is based on the Welch's test for testing whether the means across
    groups are identical assuming unequal variances across groups.

    Args:
        data: The metric data on which to perform the test of no effect.
        objective_names: The names of the objective metrics used in this analysis
            used for labeling the is_objective column in the output frame.
            If None, omit the is_objective column.
        no_effect_alpha: The confidence level at which to reject the
            sharp null of no experimental effects.

    Returns:
        A dataframe containing the results of the test of no effect per metric.

    """
    df = data.df
    df_grouped = df.groupby(["metric_name", "trial_index"])
    cols = ["metric_name", "trial_index", "p_value", "has_effect"]

    if objective_names is not None:
        cols.append("is_objective")

    df_tone = pd.DataFrame(columns=cols)

    for metric_name, trial_index in df_grouped.groups.keys():
        dfm = df_grouped.get_group((metric_name, trial_index))

        p_value, f_stat = no_effect_test_welch(
            means=list(dfm["mean"].values),
            sems=list(dfm["sem"].values),
            ns=list(dfm["n"]),
        )
        has_effect = bool(p_value < no_effect_alpha)

        d = {
            "trial_index": trial_index,
            "metric_name": metric_name,
            "p_value": p_value,
            "has_effect": has_effect,
        }
        if objective_names is not None:
            d["is_objective"] = metric_name in objective_names

        df_tone = pd.concat([df_tone, pd.DataFrame([d])], ignore_index=True)

    df_tone["trial_index"] = df_tone["trial_index"].astype(int)
    df_tone["metric_name"] = df_tone["metric_name"].astype("string")
    df_tone["has_effect"] = df_tone["has_effect"].astype(bool)

    if "is_objective" in df_tone:
        df_tone["is_objective"] = df_tone["is_objective"].astype(bool)

    return df_tone


def check_experiment_effects(
    data: Data,
    objective_names: set[str],
    status_quo_name: str = "status_quo",
    no_effect_alpha: float = 0.05,
    ci_alpha: float = 0.05,
) -> tuple[bool, list[str], pd.DataFrame]:
    """Checks whether an experiment did anything.

    Basic version without support for contexts. This checks whether a
        randomization test can show that there are any effects whatsoever.
        This test is performed independently on each metric, and the minimum
        Bonferroni adjusted p-value is returned. Additionally, returns a data
        frame showing the minimum and maximum effects consistent with the
        observed data.

    Args:
        data: The data on which to perform the check.
        objective_name: The name of the objective metric used in this analysis.
        no_effect_alpha: The confidence level at which to reject the
            sharp null of no experimental effects.
        ci_alpha: The confidence level to use to form a joint confidence
            interval for all effects.
        status_quo_name: The name of the status quo arm.

    Returns:
        effective: True if the null of no treatment effects can be rejected.
        ineffective_on_objectives: List of objectives on which the
            null of no treatment effects can be rejected.
        bounds_df: The minimum and maximum bounds on possible effects.

    """
    df = data.df
    df_grouped = df.groupby("metric_name")
    K = len(df_grouped)
    ps = []
    fx_bounds = []

    # list of objectives on which there were no effects
    # if list is non-empty, we'll show a warning
    ineffective_on_objectives = []

    for metric in df_grouped.groups.keys():
        dfm = df_grouped.get_group(metric)
        p, z, null_z = ri_test_of_no_effect(
            dfm["mean"],
            dfm["sem"],
            dfm["n"],
            dfm["arm_name"],
            status_quo_name=status_quo_name,
        )
        min_fx, max_fx = estimate_effect_bounds(
            dfm["mean"],
            dfm["sem"],
            dfm["arm_name"],
            status_quo_name=status_quo_name,
            alpha=ci_alpha,
        )
        fx_bounds.append((metric, min_fx, max_fx, p))
        ps.append(p * K)

        if metric in objective_names and p * K > no_effect_alpha:
            ineffective_on_objectives.append(metric)

    effective = np.min(ps) < no_effect_alpha
    bounds_df = pd.DataFrame(fx_bounds, columns=["metric_name", "min", "max", "p"])
    bounds_df.sort_values(by="p", inplace=True)
    return effective, ineffective_on_objectives, bounds_df


# pyre-fixme[3]: Return type must be annotated.
def ri_test_of_no_effect(
    means: list[float],
    sems: list[float],
    ns: list[int],
    names: list[str],
    status_quo_name: str,
    M: int = 10000,
):
    # pyre-fixme[9]: means has type `List[float]`; used as `ndarray`.
    means = np.array(means)
    # pyre-fixme[9]: sems has type `List[float]`; used as `ndarray`.
    sems = np.array(sems)
    # pyre-fixme[9]: ns has type `List[int]`; used as `ndarray`.
    ns = np.array(ns)
    # pyre-fixme[9]: names has type `List[str]`; used as `ndarray`.
    names = np.array(names)
    if len(names) != len(means):
        raise ValueError("Length of means and names must be equal.")
    K = len(means)
    # pyre-fixme[58]: `/` is not supported for operand types `List[int]` and `Any`.
    ps = ns / np.sum(ns)
    vars = np.power(sems, 2) * np.array(ns)
    mn = np.average(means, weights=ps)
    E_vr = np.average(vars, weights=ps**2)
    # pyre-fixme[58]: `-` is not supported for operand types `List[float]` and `Any`.
    vr_E = np.average((means - mn) ** 2, weights=ps) * K / (K - 1)
    vr = E_vr + vr_E
    z_stats = []
    p_value = 0.0
    actual_z = estimate_largest_z(means, sems, names, status_quo_name=status_quo_name)
    for _ in range(M):
        # pyre-fixme[6]: For 1st argument expected `bool` but got `List[int]`.
        b_mns = np.random.normal(loc=mn, scale=np.sqrt(vr / ns), size=K)
        # pyre-fixme[6]: For 1st argument expected `bool` but got `List[int]`.
        b_sems = np.random.normal(loc=np.sqrt(vr), scale=np.sqrt(vr / (2 * ns)), size=K)
        b_sems /= np.sqrt(ns)
        ri_arms = np.random.choice(names, K, replace=False)
        new_z = estimate_largest_z(
            b_mns, b_sems, ri_arms, status_quo_name=status_quo_name
        )
        z_stats.append(new_z)
        if new_z >= actual_z:
            p_value += 1.0 / M
    return p_value, actual_z, z_stats


def no_effect_test_welch(
    means: list[float],
    sems: list[float],
    ns: list[int],
) -> tuple[float, float]:
    r"""
    Welch's F test of no effect for the means of the arms. It tests whether the means
    across arms are identical assuming unequal variances across arms.

    The Welch's F test statistic equals:
        W* = \sum w_i (mean_i-mu_hat)^2 /(K-1) / [(1 + 2*(K-2)/(K^2-1)) \sum h_i],
    where
        mean_i, sem_i, n_i are observed means, standard errors and sample sizes
        across K groups,
        var_i = sem_i^2 * n_i is the variance of each group,
        w_i = n_i / var_i, W = sum_i w_i, mu_hat = \sum w_i * mean_i / W,
        h_i = (1-w_i/W)^2 / (n_i-1) and f = (K^2-1) / (3*\sum_i h_i).
    The value of W* is compared to an F distribution with K-1 and f degrees of freedom.

    Args:
        means: The means of the arms.
        sems: The standard errors of the arms.
        ns: The number of users per arm.

    Returns: A tuple containing
        - the p-value and
        - the test-statistic value.
    """
    means_arr = np.array(means)
    sems_arr = np.array(sems)
    ns_arr = np.array(ns)

    K = len(means_arr)

    variances = np.multiply(sems_arr**2, ns_arr)
    ws = np.divide(ns_arr, variances)

    W = np.sum(ws)

    overall_mean = np.dot(means_arr, ws) / W
    # mean square error between groups: the difference between the
    # group means and the overall weighted mean
    bg = np.dot((means_arr - overall_mean) ** 2, ws) / (K - 1)

    hs = np.divide((1.0 - ws / W) ** 2, ns_arr - 1)
    H = np.sum(hs)
    wg = 1 + 2 * (K - 2) * H / (K**2 - 1)
    dfd = (K**2 - 1) / (3 * H)

    f_stat = bg / wg
    # pyre-ignore
    p_value = 1 - scipy.stats.f.cdf(f_stat, dfn=K - 1, dfd=dfd)

    return p_value, f_stat


# pyre-fixme[3]: Return type must be annotated.
def estimate_effect_bounds(
    means: list[float],
    sems: list[float],
    names: list[str],
    status_quo_name: str | None,
    alpha: float,
):
    # pyre-fixme[9]: means has type `List[float]`; used as `ndarray[typing.Any,
    #  dtype[typing.Any]]`.
    means = np.asarray(means)
    # pyre-fixme[9]: sems has type `List[float]`; used as `ndarray[typing.Any,
    #  dtype[typing.Any]]`.
    sems = np.asarray(sems)
    # pyre-fixme[9]: names has type `List[str]`; used as `ndarray[typing.Any,
    #  dtype[typing.Any]]`.
    names = np.asarray(names)
    if status_quo_name is not None:
        # pyre-fixme[16]: `float` has no attribute `item`.
        m_c = means[names == status_quo_name].item()
        sem_c = sems[names == status_quo_name].item()
        means_t = means[names != status_quo_name]
        sems_t = sems[names != status_quo_name]
        fx, fx_sems = relativize(
            means_t=means_t, sems_t=sems_t, mean_c=m_c, sem_c=sem_c
        )
    else:
        fx, fx_sems = means, sems
    z = norm.ppf(1 - alpha / 2)
    # pyre-fixme[58]: `-` is not supported for operand types
    #  `Union[np.ndarray[typing.Any, typing.Any], typing.List[float]]` and `float`.
    # pyre-fixme[58]: `*` is not supported for operand types `float` and
    #  `Union[np.ndarray[typing.Any, typing.Any], typing.List[float]]`.
    min_fx = fx - z * fx_sems
    # pyre-fixme[58]: `+` is not supported for operand types
    #  `Union[np.ndarray[typing.Any, typing.Any], typing.List[float]]` and `float`.
    # pyre-fixme[58]: `*` is not supported for operand types `float` and
    #  `Union[np.ndarray[typing.Any, typing.Any], typing.List[float]]`.
    max_fx = fx + z * fx_sems
    return np.min(min_fx), np.max(max_fx)


# pyre-fixme[3]: Return type must be annotated.
def estimate_largest_z(
    means: list[float], sems: list[float], names: list[str], status_quo_name: str
):
    mn_sq = [mn for i, mn in enumerate(means) if names[i] == status_quo_name][0]
    sem_sq = [sm for i, sm in enumerate(sems) if names[i] == status_quo_name][0]
    arm_mns = [mn for i, mn in enumerate(means) if names[i] != status_quo_name]
    arm_sems = [sm for i, sm in enumerate(sems) if names[i] != status_quo_name]
    effects = [mn - mn_sq for mn in arm_mns]
    effect_sems = [np.sqrt(sem**2 + sem_sq**2) for sem in arm_sems]
    return np.max(np.abs(effects) / effect_sems)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from collections.abc import Sequence
from functools import reduce

from ax.core.experiment import Experiment
from ax.early_stopping.strategies.base import BaseEarlyStoppingStrategy
from ax.exceptions.core import UserInputError
from ax.generation_strategy.generation_node import GenerationNode


class LogicalEarlyStoppingStrategy(BaseEarlyStoppingStrategy):
    def __init__(
        self,
        left: BaseEarlyStoppingStrategy,
        right: BaseEarlyStoppingStrategy,
    ) -> None:
        super().__init__()

        self.left = left
        self.right = right


class AndEarlyStoppingStrategy(LogicalEarlyStoppingStrategy):
    def should_stop_trials_early(
        self,
        trial_indices: set[int],
        experiment: Experiment,
        current_node: GenerationNode | None = None,
    ) -> dict[int, str | None]:
        left = self.left.should_stop_trials_early(
            trial_indices=trial_indices,
            experiment=experiment,
            current_node=current_node,
        )
        right = self.right.should_stop_trials_early(
            trial_indices=trial_indices,
            experiment=experiment,
            current_node=current_node,
        )
        return {
            trial: f"{left[trial]}, {right[trial]}" for trial in left if trial in right
        }


class OrEarlyStoppingStrategy(LogicalEarlyStoppingStrategy):
    @classmethod
    def from_early_stopping_strategies(
        cls,
        strategies: Sequence[BaseEarlyStoppingStrategy],
    ) -> BaseEarlyStoppingStrategy:
        if len(strategies) < 1:
            raise UserInputError("strategies must not be empty")

        return reduce(
            lambda left, right: OrEarlyStoppingStrategy(left=left, right=right),
            strategies[1:],
            strategies[0],
        )

    def should_stop_trials_early(
        self,
        trial_indices: set[int],
        experiment: Experiment,
        current_node: GenerationNode | None = None,
    ) -> dict[int, str | None]:
        return {
            **self.left.should_stop_trials_early(
                trial_indices=trial_indices,
                experiment=experiment,
                current_node=current_node,
            ),
            **self.right.should_stop_trials_early(
                trial_indices=trial_indices,
                experiment=experiment,
                current_node=current_node,
            ),
        }



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from collections.abc import Iterable, Sequence
from typing import Any

from ax.core.arm import Arm
from ax.core.base_trial import BaseTrial, TrialStatus
from ax.core.data import Data
from ax.core.experiment import DataType, Experiment
from ax.core.metric import Metric, MetricFetchResult
from ax.core.optimization_config import OptimizationConfig
from ax.core.runner import Runner
from ax.core.search_space import SearchSpace
from ax.utils.common.docutils import copy_doc
from pyre_extensions import none_throws


class MultiTypeExperiment(Experiment):
    """Class for experiment with multiple trial types.

    A canonical use case for this is tuning a large production system
    with limited evaluation budget and a simulator which approximates
    evaluations on the main system. Trial deployment and data fetching
    is separate for the two systems, but the final data is combined and
    fed into multi-task models.

    See the Multi-Task Modeling tutorial for more details.

    Attributes:
        name: Name of the experiment.
        description: Description of the experiment.
    """

    def __init__(
        self,
        name: str,
        search_space: SearchSpace,
        default_trial_type: str,
        default_runner: Runner | None,
        optimization_config: OptimizationConfig | None = None,
        tracking_metrics: list[Metric] | None = None,
        status_quo: Arm | None = None,
        description: str | None = None,
        is_test: bool = False,
        experiment_type: str | None = None,
        properties: dict[str, Any] | None = None,
        default_data_type: DataType | None = None,
    ) -> None:
        """Inits Experiment.

        Args:
            name: Name of the experiment.
            search_space: Search space of the experiment.
            default_trial_type: Default type for trials on this experiment.
            default_runner: Default runner for trials of the default type.
            optimization_config: Optimization config of the experiment.
            tracking_metrics: Additional tracking metrics not used for optimization.
                These are associated with the default trial type.
            runner: Default runner used for trials on this experiment.
            status_quo: Arm representing existing "control" arm.
            description: Description of the experiment.
            is_test: Convenience metadata tracker for the user to mark test experiments.
            experiment_type: The class of experiments this one belongs to.
            properties: Dictionary of this experiment's properties.
            default_data_type: Enum representing the data type this experiment uses.
        """

        self._default_trial_type = default_trial_type

        # Map from trial type to default runner of that type
        self._trial_type_to_runner: dict[str, Runner | None] = {
            default_trial_type: default_runner
        }

        # Specifies which trial type each metric belongs to
        self._metric_to_trial_type: dict[str, str] = {}

        # Maps certain metric names to a canonical name. Useful for ancillary trial
        # types' metrics, to specify which primary metrics they correspond to
        # (e.g. 'comment_prediction' => 'comment')
        self._metric_to_canonical_name: dict[str, str] = {}

        # call super.__init__() after defining fields above, because we need
        # them to be populated before optimization config is set
        super().__init__(
            name=name,
            search_space=search_space,
            optimization_config=optimization_config,
            status_quo=status_quo,
            description=description,
            is_test=is_test,
            experiment_type=experiment_type,
            properties=properties,
            default_data_type=default_data_type,
            tracking_metrics=tracking_metrics,
        )

    def add_trial_type(self, trial_type: str, runner: Runner) -> "MultiTypeExperiment":
        """Add a new trial_type to be supported by this experiment.

        Args:
            trial_type: The new trial_type to be added.
            runner: The default runner for trials of this type.
        """
        if self.supports_trial_type(trial_type):
            raise ValueError(f"Experiment already contains trial_type `{trial_type}`")

        self._trial_type_to_runner[trial_type] = runner
        return self

    # pyre-fixme [56]: Pyre was not able to infer the type of the decorator
    # `Experiment.optimization_config.setter`.
    @Experiment.optimization_config.setter
    def optimization_config(self, optimization_config: OptimizationConfig) -> None:
        # pyre-fixme [16]: `Optional` has no attribute `fset`.
        Experiment.optimization_config.fset(self, optimization_config)
        for metric_name in optimization_config.metrics.keys():
            # Optimization config metrics are required to be the default trial type
            # currently. TODO: remove that restriction (T202797235)
            self._metric_to_trial_type[metric_name] = none_throws(
                self.default_trial_type
            )

    def update_runner(self, trial_type: str, runner: Runner) -> "MultiTypeExperiment":
        """Update the default runner for an existing trial_type.

        Args:
            trial_type: The new trial_type to be added.
            runner: The new runner for trials of this type.
        """
        if not self.supports_trial_type(trial_type):
            raise ValueError(f"Experiment does not contain trial_type `{trial_type}`")

        self._trial_type_to_runner[trial_type] = runner
        return self

    # pyre-fixme[14]: `add_tracking_metric` overrides method defined in `Experiment`
    #  inconsistently.
    def add_tracking_metric(
        self,
        metric: Metric,
        trial_type: str | None = None,
        canonical_name: str | None = None,
    ) -> "MultiTypeExperiment":
        """Add a new metric to the experiment.

        Args:
            metric: The metric to add.
            trial_type: The trial type for which this metric is used.
            canonical_name: The default metric for which this metric is a proxy.
        """
        if trial_type is None:
            trial_type = self._default_trial_type
        if not self.supports_trial_type(trial_type):
            raise ValueError(f"`{trial_type}` is not a supported trial type.")

        super().add_tracking_metric(metric)
        self._metric_to_trial_type[metric.name] = trial_type
        if canonical_name is not None:
            self._metric_to_canonical_name[metric.name] = canonical_name
        return self

    def add_tracking_metrics(
        self,
        metrics: list[Metric],
        metrics_to_trial_types: dict[str, str] | None = None,
        canonical_names: dict[str, str] | None = None,
    ) -> Experiment:
        """Add a list of new metrics to the experiment.

        If any of the metrics are already defined on the experiment,
        we raise an error and don't add any of them to the experiment

        Args:
            metrics: Metrics to be added.
            metrics_to_trial_types: The mapping from metric names to corresponding
                trial types for each metric. If provided, the metrics will be
                added to their trial types. If not provided, then the default
                trial type will be used.
            canonical_names: A mapping of metric names to their
                canonical names(The default metrics for which the metrics are
                proxies.)

        Returns:
            The experiment with the added metrics.
        """
        metrics_to_trial_types = metrics_to_trial_types or {}
        canonical_name = None
        for metric in metrics:
            if canonical_names is not None:
                canonical_name = none_throws(canonical_names).get(metric.name, None)

            self.add_tracking_metric(
                metric=metric,
                trial_type=metrics_to_trial_types.get(
                    metric.name, self._default_trial_type
                ),
                canonical_name=canonical_name,
            )
        return self

    # pyre-fixme[14]: `update_tracking_metric` overrides method defined in
    #  `Experiment` inconsistently.
    def update_tracking_metric(
        self, metric: Metric, trial_type: str, canonical_name: str | None = None
    ) -> "MultiTypeExperiment":
        """Update an existing metric on the experiment.

        Args:
            metric: The metric to add.
            trial_type: The trial type for which this metric is used.
            canonical_name: The default metric for which this metric is a proxy.
        """
        oc = self.optimization_config
        oc_metrics = oc.metrics if oc else []
        if metric.name in oc_metrics and trial_type != self._default_trial_type:
            raise ValueError(
                f"Metric `{metric.name}` must remain a `{self._default_trial_type}` "
                "metric because it is part of the optimization_config."
            )
        elif not self.supports_trial_type(trial_type):
            raise ValueError(f"`{trial_type}` is not a supported trial type.")

        super().update_tracking_metric(metric)
        self._metric_to_trial_type[metric.name] = trial_type
        if canonical_name is not None:
            self._metric_to_canonical_name[metric.name] = canonical_name
        return self

    @copy_doc(Experiment.remove_tracking_metric)
    def remove_tracking_metric(self, metric_name: str) -> "MultiTypeExperiment":
        if metric_name not in self._tracking_metrics:
            raise ValueError(f"Metric `{metric_name}` doesn't exist on experiment.")

        # Required fields
        del self._tracking_metrics[metric_name]
        del self._metric_to_trial_type[metric_name]

        # Optional
        if metric_name in self._metric_to_canonical_name:
            del self._metric_to_canonical_name[metric_name]
        return self

    @copy_doc(Experiment.fetch_data)
    def fetch_data(
        self,
        trial_indices: Iterable[int] | None = None,
        metrics: list[Metric] | None = None,
        combine_with_last_data: bool = False,
        overwrite_existing_data: bool = False,
        **kwargs: Any,
    ) -> Data:
        # TODO: make this more efficient for fetching
        # data for multiple trials of the same type
        # by overriding Experiment._lookup_or_fetch_trials_results
        return self.default_data_constructor.from_multiple_data(
            [
                (
                    trial.fetch_data(**kwargs, metrics=metrics)
                    if trial.status.expecting_data
                    else Data()
                )
                for trial in self.trials.values()
            ]
        )

    @copy_doc(Experiment._fetch_trial_data)
    def _fetch_trial_data(
        self, trial_index: int, metrics: list[Metric] | None = None, **kwargs: Any
    ) -> dict[str, MetricFetchResult]:
        trial = self.trials[trial_index]
        metrics = [
            metric
            for metric in (metrics or self.metrics.values())
            if self.metric_to_trial_type[metric.name] == trial.trial_type
        ]
        # Invoke parent's fetch method using only metrics for this trial_type
        return super()._fetch_trial_data(trial.index, metrics=metrics, **kwargs)

    @property
    def default_trials(self) -> set[int]:
        """Return the indicies for trials of the default type."""
        return {
            idx
            for idx, trial in self.trials.items()
            if trial.trial_type == self.default_trial_type
        }

    @property
    def metric_to_trial_type(self) -> dict[str, str]:
        """Map metrics to trial types.

        Adds in default trial type for OC metrics to custom defined trial types..
        """
        opt_config_types = {
            metric_name: self.default_trial_type
            for metric_name in self.optimization_config.metrics.keys()
        }
        return {**opt_config_types, **self._metric_to_trial_type}

    # -- Overridden functions from Base Experiment Class --
    @property
    def default_trial_type(self) -> str | None:
        """Default trial type assigned to trials in this experiment."""
        return self._default_trial_type

    def runner_for_trial(self, trial: BaseTrial) -> Runner | None:
        """The default runner to use for a given trial.

        Looks up the appropriate runner for this trial type in the trial_type_to_runner.
        """
        return (
            trial._runner
            if trial._runner
            else self.runner_for_trial_type(trial_type=none_throws(trial.trial_type))
        )

    def runner_for_trial_type(self, trial_type: str) -> Runner | None:
        """The default runner to use for a given trial type.

        Looks up the appropriate runner for this trial type in the trial_type_to_runner.
        """
        if not self.supports_trial_type(trial_type):
            raise ValueError(f"Trial type `{trial_type}` is not supported.")
        return self._trial_type_to_runner[trial_type]

    def metrics_for_trial_type(self, trial_type: str) -> list[Metric]:
        """The default runner to use for a given trial type.

        Looks up the appropriate runner for this trial type in the trial_type_to_runner.
        """
        if not self.supports_trial_type(trial_type):
            raise ValueError(f"Trial type `{trial_type}` is not supported.")
        return [
            self.metrics[metric_name]
            for metric_name, metric_trial_type in self._metric_to_trial_type.items()
            if metric_trial_type == trial_type
        ]

    def supports_trial_type(self, trial_type: str | None) -> bool:
        """Whether this experiment allows trials of the given type.

        Only trial types defined in the trial_type_to_runner are allowed.
        """
        return trial_type in self._trial_type_to_runner.keys()

    def reset_runners(self, runner: Runner) -> None:
        raise NotImplementedError(
            "MultiTypeExperiment does not support resetting all runners."
        )


def filter_trials_by_type(
    trials: Sequence[BaseTrial], trial_type: str | None
) -> list[BaseTrial]:
    """Filter trials by trial type if provided.

    This filters trials by trial type if the experiment is a
    MultiTypeExperiment.

    Args:
        trials: Trials to filter.

    Returns:
        Filtered trials.
    """
    if trial_type is not None:
        return [t for t in trials if t.trial_type == trial_type]
    return list(trials)


def get_trial_indices_for_statuses(
    experiment: Experiment, statuses: set[TrialStatus], trial_type: str | None = None
) -> set[int]:
    """Get trial indices for a set of statuses.

    Args:
        statuses: Set of statuses to get trial indices for.

    Returns:
        Set of trial indices for the given statuses.
    """
    return {
        i
        for i, t in experiment.trials.items()
        if (t.status in statuses)
        and (
            (trial_type is None)
            or ((trial_type is not None) and (t.trial_type == trial_type))
        )
    }



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from copy import deepcopy
from unittest.mock import patch

import numpy as np
from ax.adapter.base import DataLoaderConfig
from ax.adapter.data_utils import ExperimentData, extract_experiment_data
from ax.adapter.transforms.cast import Cast
from ax.core.observation import Observation, ObservationData, ObservationFeatures
from ax.core.parameter import (
    ChoiceParameter,
    FixedParameter,
    ParameterType,
    RangeParameter,
)
from ax.core.search_space import HierarchicalSearchSpace, SearchSpace
from ax.exceptions.core import UserInputError
from ax.utils.common.constants import Keys
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import (
    get_branin_experiment_with_timestamp_map_metric,
    get_experiment_with_observations,
    get_hierarchical_search_space,
)
from pandas import DataFrame
from pandas.testing import assert_frame_equal


class CastTransformTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.search_space = SearchSpace(
            parameters=[
                RangeParameter(
                    "a", lower=1.0, upper=5.0, parameter_type=ParameterType.FLOAT
                ),
                RangeParameter(
                    "b",
                    lower=1.0,
                    upper=5.0,
                    digits=2,
                    parameter_type=ParameterType.FLOAT,
                ),
                ChoiceParameter(
                    "c", parameter_type=ParameterType.STRING, values=["a", "b", "c"]
                ),
                FixedParameter(name="d", parameter_type=ParameterType.INT, value=2),
            ],
            parameter_constraints=[],
        )
        self.t = Cast(search_space=self.search_space, observations=[])
        self.hss = get_hierarchical_search_space()
        self.t_hss = Cast(search_space=self.hss, observations=[])
        self.obs_feats_hss = ObservationFeatures(
            parameters={
                "model": "Linear",
                "learning_rate": 0.01,
                "l2_reg_weight": 0.0001,
                "num_boost_rounds": 12,
            },
            trial_index=9,
            metadata=None,
        )
        self.obs_feats_hss_2 = ObservationFeatures(
            parameters={
                "model": "XGBoost",
                "learning_rate": 0.01,
                "l2_reg_weight": 0.0001,
                "num_boost_rounds": 12,
            },
            trial_index=10,
            metadata=None,
        )
        self.obs_data = ObservationData(
            metric_names=["m1"], means=np.array([1.0]), covariance=np.array([[1.0]])
        )

    def test_invalid_config(self) -> None:
        with self.assertRaisesRegex(UserInputError, "Unexpected config"):
            Cast(search_space=self.search_space, config={"flatten_hs": "foo"})

    def test_transform_observations_and_features(self) -> None:
        # Verify running the transform on already-casted features does nothing
        observation_features = [
            ObservationFeatures(parameters={"a": 1.2345, "b": 2.34, "c": "a", "d": 2})
        ]
        obs_ft2 = deepcopy(observation_features)
        obs_ft2 = self.t.transform_observation_features(obs_ft2)
        self.assertEqual(obs_ft2, observation_features)
        obs_ft2 = self.t.untransform_observation_features(obs_ft2)
        self.assertEqual(obs_ft2, observation_features)

        # Test with transform_observations.
        obs = Observation(features=obs_ft2[0], data=self.obs_data, arm_name="arm")
        (tf_obs,) = self.t.transform_observations([obs])
        self.assertEqual(tf_obs.features, observation_features[0])
        self.assertEqual(tf_obs.data, self.obs_data)
        self.assertEqual(tf_obs.arm_name, "arm")

        # Check that the transform casts the parameter values when necessary.
        observation_features = [
            ObservationFeatures(parameters={"a": 1, "b": 2, "c": "a", "d": 2.1})
        ]
        expected = [
            ObservationFeatures(parameters={"a": 1.0, "b": 2.0, "c": "a", "d": 2})
        ]
        self.assertEqual(
            self.t.transform_observation_features(
                observation_features=observation_features
            ),
            expected,
        )

    def test_untransform_observation_features(self) -> None:
        # Verify running the transform on uncasted values properly converts them
        # (e.g. typing, rounding)
        observation_features = [
            ObservationFeatures(parameters={"a": 1, "b": 2.3466789, "c": "a", "d": 2.0})
        ]
        observation_features = self.t.untransform_observation_features(
            observation_features
        )
        self.assertEqual(
            observation_features,
            [ObservationFeatures(parameters={"a": 1.0, "b": 2.35, "c": "a", "d": 2})],
        )

    def test_flatten_hss_setting(self) -> None:
        t = Cast(search_space=self.hss, observations=[])
        self.assertTrue(t.flatten_hss)
        t = Cast(search_space=self.hss, config={"flatten_hss": False}, observations=[])
        self.assertFalse(t.flatten_hss)
        self.assertFalse(self.t.flatten_hss)  # `self.t` does not have HSS
        self.assertTrue(self.t_hss.flatten_hss)  # `self.t_hss` does have HSS

    def test_transform_search_space_HSS(self) -> None:
        with patch.object(
            self.hss, "flatten", wraps=self.hss.flatten
        ) as mock_hss_flatten:
            flattened_search_space = self.t_hss.transform_search_space(
                search_space=self.hss
            )
        mock_hss_flatten.assert_called_once()
        self.assertIsNot(flattened_search_space, self.hss)
        self.assertFalse(isinstance(flattened_search_space, HierarchicalSearchSpace))

    def test_transform_observation_features_HSS(self) -> None:
        # Untransform the observation features first to cast them and
        # save their full parameterization in metadata.
        obs_feats = self.t_hss.untransform_observation_features(
            observation_features=[self.obs_feats_hss]
        )
        with patch.object(
            self.t_hss.search_space,
            "flatten_observation_features",
            wraps=self.t_hss.search_space.flatten_observation_features,  # pyre-ignore
        ) as mock_flatten_obsf:
            transformed_obs_feats = self.t_hss.transform_observation_features(
                observation_features=obs_feats
            )
        mock_flatten_obsf.assert_called_once()

        for obsf in transformed_obs_feats:
            # Check that transformed obs feats have all the parameters
            for p_name in self.t_hss.search_space.parameters:
                self.assertIn(p_name, obsf.parameters)
            # Check that full parameterization is recorded in metadata
            self.assertEqual(
                # pyre-fixme[16]: Optional type has no attribute `get`.
                obsf.metadata.get(Keys.FULL_PARAMETERIZATION),
                self.obs_feats_hss.parameters,
            )

        # Perform one more roundtrip so parameterizations are cast to HSS.
        obs_feats = self.t_hss.untransform_observation_features(
            observation_features=transformed_obs_feats
        )
        new_transformed_obs_feats = self.t_hss.transform_observation_features(
            observation_features=obs_feats
        )
        for obsf in new_transformed_obs_feats:
            # Check that transformed obs feats have all the parameters
            for p_name in self.t_hss.search_space.parameters:
                self.assertIn(p_name, obsf.parameters)
            # Check that full parameterization is recorded in metadata
            self.assertEqual(
                obsf.metadata.get(Keys.FULL_PARAMETERIZATION),
                self.obs_feats_hss.parameters,
            )

    def test_transform_observation_features_HSS_dummy_values_settings(self) -> None:
        t = Cast(
            search_space=self.hss,
            config={
                "inject_dummy_values_to_complete_flat_parameterization": True,
                "use_random_dummy_values": True,
            },
            observations=[],
        )
        self.assertTrue(t.inject_dummy_values_to_complete_flat_parameterization)
        with patch.object(
            t.search_space,
            "flatten_observation_features",
            wraps=t.search_space.flatten_observation_features,  # pyre-ignore
        ) as mock_flatten_obsf:
            t.transform_observation_features(observation_features=[self.obs_feats_hss])
        mock_flatten_obsf.assert_called_once()
        self.assertTrue(
            mock_flatten_obsf.call_args.kwargs[
                "inject_dummy_values_to_complete_flat_parameterization"
            ]
        )
        self.assertTrue(mock_flatten_obsf.call_args.kwargs["use_random_dummy_values"])

    def test_untransform_observation_features_HSS(self) -> None:
        # Test transformation in one subtree of HSS.
        with patch.object(
            self.t_hss.search_space,
            "cast_observation_features",
            wraps=self.t_hss.search_space.cast_observation_features,  # pyre-ignore
        ) as mock_cast_obsf:
            obs_feats = self.t_hss.untransform_observation_features(
                observation_features=[self.obs_feats_hss]
            )
        mock_cast_obsf.assert_called_once()

        self.assertEqual(len(obs_feats), 1)
        obsf = obs_feats[0]
        self.assertEqual(
            obsf.parameters,
            {
                "model": "Linear",
                "learning_rate": 0.01,
                "l2_reg_weight": 0.0001,
            },
        )
        self.assertEqual(
            # pyre-fixme[16]: Optional type has no attribute `get`.
            obsf.metadata.get(Keys.FULL_PARAMETERIZATION),
            self.obs_feats_hss.parameters,
        )

        # Test transformation in other subtree of HSS.
        obs_feats_2 = self.t_hss.untransform_observation_features(
            observation_features=[self.obs_feats_hss_2]
        )
        self.assertEqual(len(obs_feats_2), 1)
        obsf = obs_feats_2[0]
        self.assertEqual(
            obsf.parameters,
            {
                "model": "XGBoost",
                "num_boost_rounds": 12,
            },
        )
        self.assertEqual(
            obsf.metadata.get(Keys.FULL_PARAMETERIZATION),
            self.obs_feats_hss_2.parameters,
        )

    def test_cast_parameter_type_and_none(self) -> None:
        # This test covers removal of observations with Nones, casting
        # to correct parameter type and rounding to digits for RangeParameters.
        search_space = SearchSpace(
            parameters=[
                ChoiceParameter(
                    name="choice",
                    parameter_type=ParameterType.STRING,
                    values=["1", "2", "3"],
                ),
                RangeParameter(
                    name="range",
                    parameter_type=ParameterType.FLOAT,
                    lower=0.0,
                    upper=5.0,
                    digits=1,
                ),
            ]
        )
        t = Cast(search_space=search_space)
        obs_features = [
            ObservationFeatures(parameters={"choice": None, "range": 5.0}),
            ObservationFeatures(parameters={"choice": 1, "range": 3}),
            ObservationFeatures(parameters={"choice": "2", "range": 3.567}),
        ]
        observations = [
            Observation(
                features=ft.clone(), data=deepcopy(self.obs_data), arm_name=f"{i}"
            )
            for i, ft in enumerate(obs_features)
        ]
        tf_obs_features = t.transform_observation_features(
            observation_features=obs_features
        )
        self.assertEqual(
            tf_obs_features,
            [
                ObservationFeatures(parameters={"choice": "1", "range": 3.0}),
                ObservationFeatures(parameters={"choice": "2", "range": 3.6}),
            ],
        )
        tf_observations = t.transform_observations(observations)
        expected = [
            Observation(
                features=ObservationFeatures(parameters={"choice": "1", "range": 3.0}),
                data=self.obs_data,
                arm_name="1",
            ),
            Observation(
                features=ObservationFeatures(parameters={"choice": "2", "range": 3.6}),
                data=self.obs_data,
                arm_name="2",
            ),
        ]
        self.assertEqual(tf_observations, expected)

    def test_transform_experiment_data_flatten(self) -> None:
        # Tests for flattening of hierarchical parameterizations.
        columns = [
            "model",
            "learning_rate",
            "l2_reg_weight",
            "num_boost_rounds",
            "metadata",
        ]
        arm_data = DataFrame.from_dict(  # Same data used in `setUp`.
            {
                (0, "0_0"): {
                    "model": "Linear",
                    "learning_rate": 0.01,
                    "l2_reg_weight": 0.0001,
                    "metadata": {
                        Keys.FULL_PARAMETERIZATION: {
                            "model": "Linear",
                            "learning_rate": 0.01,
                            "l2_reg_weight": 0.0001,
                            "num_boost_rounds": 12,
                        }
                    },
                },
                (1, "1_0"): {
                    "model": "XGBoost",
                    "num_boost_rounds": 12,
                    "metadata": {
                        Keys.FULL_PARAMETERIZATION: {
                            "model": "XGBoost",
                            "learning_rate": 0.01,
                            "l2_reg_weight": 0.0001,
                            "num_boost_rounds": 12,
                        }
                    },
                },
            },
            orient="index",
            columns=columns,
        )
        arm_data.index.names = ["trial_index", "arm_name"]
        experiment_data = ExperimentData(
            arm_data=arm_data, observation_data=DataFrame()
        )
        transformed = self.t_hss.transform_experiment_data(
            experiment_data=experiment_data
        )
        expected_arm_data = DataFrame.from_dict(
            {
                (0, "0_0"): {
                    "model": "Linear",
                    "learning_rate": 0.01,
                    "l2_reg_weight": 0.0001,
                    "num_boost_rounds": 12,
                    "metadata": {
                        Keys.FULL_PARAMETERIZATION: {
                            "model": "Linear",
                            "learning_rate": 0.01,
                            "l2_reg_weight": 0.0001,
                            "num_boost_rounds": 12,
                        }
                    },
                },
                (1, "1_0"): {
                    "model": "XGBoost",
                    "learning_rate": 0.01,
                    "l2_reg_weight": 0.0001,
                    "num_boost_rounds": 12,
                    "metadata": {
                        Keys.FULL_PARAMETERIZATION: {
                            "model": "XGBoost",
                            "learning_rate": 0.01,
                            "l2_reg_weight": 0.0001,
                            "num_boost_rounds": 12,
                        }
                    },
                },
            },
            orient="index",
            columns=columns,
        )
        expected_arm_data.index.names = ["trial_index", "arm_name"]
        assert_frame_equal(transformed.arm_data, expected_arm_data)

    def test_transform_experiment_data_flatten_with_missing_columns(self) -> None:
        columns = ["model", "learning_rate", "l2_reg_weight", "metadata"]
        arm_data = (
            DataFrame.from_dict(  # Data intentionally missing `num_boost_rounds`.
                {
                    (0, "0_0"): {
                        "model": "Linear",
                        "learning_rate": 0.01,
                        "l2_reg_weight": 0.0001,
                        "metadata": {
                            Keys.FULL_PARAMETERIZATION: {
                                "model": "Linear",
                                "learning_rate": 0.01,
                                "l2_reg_weight": 0.0001,
                            }
                        },
                    }
                },
                orient="index",
                columns=columns,
            )
        )
        arm_data.index.names = ["trial_index", "arm_name"]
        experiment_data = ExperimentData(
            arm_data=arm_data, observation_data=DataFrame()
        )
        transformed = self.t_hss.transform_experiment_data(
            experiment_data=experiment_data
        )
        expected_columns = set(columns + ["num_boost_rounds"])
        self.assertEqual(set(transformed.arm_data.columns), expected_columns)
        # Test with empty DF w/ missing columns.
        arm_data = arm_data.iloc[:0]
        arm_data.index.names = ["trial_index", "arm_name"]
        experiment_data = ExperimentData(
            arm_data=arm_data, observation_data=DataFrame()
        )
        transformed = self.t_hss.transform_experiment_data(
            experiment_data=experiment_data
        )
        self.assertEqual(set(transformed.arm_data.columns), expected_columns)

    def test_transform_experiment_data_cast(self) -> None:
        # Test for casting to the correct data type and dropping of Nones.
        experiment = get_experiment_with_observations(
            observations=[[0.0], [1.0], [2.0]],
            parameterizations=[
                {"x": 1, "y": None},
                {"x": 2, "y": 2.0},
                {"x": 3, "y": 3},
            ],
        )
        experiment_data = extract_experiment_data(
            experiment=experiment, data_loader_config=DataLoaderConfig()
        )
        transformed = Cast(
            search_space=experiment.search_space
        ).transform_experiment_data(experiment_data=deepcopy(experiment_data))
        # Arm data should drop row 0 and cast to float.
        expected_arm_data = (
            experiment_data.arm_data.copy(deep=True)
            .iloc[[1, 2]]
            .astype({"x": float, "y": float})
        )
        assert_frame_equal(transformed.arm_data, expected_arm_data)
        # Observation data should drop row 0.
        expected_obs_data = experiment_data.observation_data.copy(deep=True).iloc[
            [1, 2]
        ]
        assert_frame_equal(transformed.observation_data, expected_obs_data)

    def test_transform_experiment_data_cast_map_data(self) -> None:
        # Check that indexing for removal of NaNs works correctly with MapData.
        experiment = get_branin_experiment_with_timestamp_map_metric(
            with_trials_and_data=True
        )
        # Add some data for the last trial as well.
        experiment.fetch_data()
        # Update the last trial to mark parameterization as None.
        experiment.trials[2].arms[0]._parameters["x1"] = None

        experiment_data = extract_experiment_data(
            experiment=experiment,
            data_loader_config=DataLoaderConfig(
                fit_only_completed_map_metrics=False,
                latest_rows_per_group=None,
            ),
        )
        transformed_data = Cast(
            search_space=experiment.search_space
        ).transform_experiment_data(experiment_data=deepcopy(experiment_data))
        # Arm data should only include first three rows.
        assert_frame_equal(transformed_data.arm_data, experiment_data.arm_data.iloc[:2])
        # Observation data should include all but rows for last trial.
        assert_frame_equal(
            transformed_data.observation_data,
            experiment_data.observation_data.iloc[:-2],
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from dataclasses import dataclass

from ax.early_stopping.strategies.base import BaseEarlyStoppingStrategy

from ax.generation_strategy.generation_strategy import GenerationStrategy
from ax.utils.common.base import Base


@dataclass(kw_only=True)
class BenchmarkMethod(Base):
    """Benchmark method, represented in terms of Ax generation strategy (which tells us
    which models to use when) and Orchestrator options (which tell us extra execution
    information like maximum parallelism, early stopping configuration, etc.).

    Args:
        name: String description.
        generation_strategy: The `GenerationStrategy` to use.
        timeout_hours: Number of hours after which to stop a benchmark
            replication.
        distribute_replications: Indicates whether the replications should be
            run in a distributed manner. Ax itself does not use this attribute.
        batch_size: Number of arms per trial. If greater than 1, trials are
            ``BatchTrial``s; otherwise, they are ``Trial``s. Defaults to 1. This
            and the following arguments are passed to ``OrchestratorOptions``.
        run_trials_in_batches: Passed to ``OrchestratorOptions``.
        max_pending_trials: Passed to ``OrchestratorOptions``.
    """

    name: str = "DEFAULT"
    generation_strategy: GenerationStrategy

    timeout_hours: float = 4.0
    distribute_replications: bool = False

    batch_size: int | None = 1
    run_trials_in_batches: bool = False
    max_pending_trials: int = 1
    early_stopping_strategy: BaseEarlyStoppingStrategy | None = None

    def __post_init__(self) -> None:
        if self.name == "DEFAULT":
            self.name = self.generation_strategy.name
